# Builtins
from datetime import timedelta, timezone
import datetime as dt
from typing import Any, List, Tuple
import math

# External libraries
from finta import TA
import numpy as np
import pandas as pd

from harvest.utils import *
from harvest.plugin._base import Plugin

"""
Methods that perform an action should follow this naming convention:
    [action]_[target]_[returns]
    - action: The action taken by the method, such as 'get', 'buy'
    - target: The entity the method operates on, such as 'stock', 'option', 'account'
    - returns: What the method returns, such as 'list', 'price'

Algo class is the main interface between users and the program, so
remember the following guidelines for handling inputs and outputs:
-   Date inputs can be a string, a naive DateTime object, or a Timestamp object.
    These must be converted to UTC timezone DateTime objects using convert_input_to_datetime()
    before being passed onto other parts of the code. The same applies for timedelta inputs.
-   Conversely, date outputs must be converted to a naive DateTime object set to the user's timezone.
"""


class BaseAlgo:
    """
    The BaseAlgo class is an abstract class defining the interface for users to
    track assets, monitor their accounts, and place orders.
    It also provides function for technical analysis.
    """

    def __init__(self):
        self.trader = None
        self.interval = None
        self.aggregations = None
        self.watchlist = []

    def config(self):
        """
        This method is called before any other methods (except for __init__),
        and initializes parameters for this class.
        -  interval: The interval to run the algorithm.
        -  aggregations: Intervals to aggregate data.
        -  watchlist: List of assets this algorithm tracks.
        Any parameters set to None or an empty List will fall back to respective paramters set in the Trader class.
        """
        self.interval = None
        self.aggregations = None
        self.watchlist = []

    def setup(self):
        """
        Method called right before algorithm begins.
        """
        pass

    def main(self):
        """
        Main method to run the algorithm.
        """
        pass

    def add_plugin(self, plugin: Plugin):
        """
        Adds a plugin to the algorithm.
        """
        value = getattr(self, plugin.name, None)
        if value is None:
            setattr(self, plugin.name, plugin)
        else:
            debugger.error(
                f"Plugin name is already in use! {plugin.name} points to {value}."
            )

    ############ Functions interfacing with broker through the trader #################

    def buy(
        self,
        symbol: str = None,
        quantity: int = None,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        """Buys the specified asset.

        When called, a limit buy order is placed with a limit
        price 5% higher than the current price.

        :param str? symbol: Symbol of the asset to buy. defaults to first symbol in watchlist
        :param float? quantity: Quantity of asset to buy. defaults to buys as many as possible
        :param str? in_force: Duration the order is in force. '{gtc}' or '{gtd}'. defaults to 'gtc'
        :param str? extended: Whether to trade in extended hours or not. defaults to False
        :returns: The following Python dictionary

            - id: str, ID of order
            - symbol: str, symbol of asset

        :raises Exception: There is an error in the order process.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        if quantity is None:
            quantity = self.get_asset_max_quantity(symbol)

        debugger.debug(f"Algo BUY: {symbol}, {quantity}")
        return self.trader.buy(symbol, quantity, in_force, extended)

    def sell(
        self,
        symbol: str = None,
        quantity: int = None,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        """Sells the specified asset.

        When called, a limit sell order is placed with a limit
        price 5% lower than the current price.

        :param str? symbol:    Symbol of the asset to sell. defaults to first symbol in watchlist
        :param float? quantity:  Quantity of asset to sell defaults to sells all
        :param str? in_force:  Duration the order is in force. '{gtc}' or '{gtd}'. defaults to 'gtc'
        :param str? extended:  Whether to trade in extended hours or not. defaults to False
        :returns: A dictionary with the following keys:

            - id: str, ID of order
            - symbol: str, symbol of asset

        :raises Exception: There is an error in the order process.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        if quantity is None:
            quantity = self.get_asset_quantity(symbol)

        debugger.debug(f"Algo SELL: {symbol}, {quantity}")
        return self.trader.sell(symbol, quantity, in_force, extended)

    def sell_all_options(self, symbol: str = None, in_force: str = "gtc"):
        """Sells all options of a stock

        :param str? symbol: symbol of stock. defaults to first symbol in watchlist
        :returns: A list of dictionaries with the following keys:

            - id: str, ID of order
            - symbol: str, symbol of asset
        """
        if symbol is None:
            symbol = self.watchlist[0]

        symbols = [
            s["symbol"]
            for s in self.get_account_option_positions()
            if s["base_symbol"] == symbol
        ]
        ret = []
        for s in symbols:
            debugger.debug(f"Algo SELL OPTION: {s}")
            quantity = self.get_asset_quantity(s)
            ret.append(self.trader.sell(s, quantity, in_force, True))

        return ret

    # def buy_option(self, symbol: str, quantity: int = None, in_force: str = "gtc"):
    #     """Buys the specified option.

    #     When called, a limit buy order is placed with a limit
    #     price 5% higher than the current price.

    #     :param str symbol:    Symbol of the asset to buy, in {OCC} format.
    #     :param float? quantity:  Quantity of asset to buy. defaults to buys as many as possible
    #     :param str? in_force:  Duration the order is in force. '{gtc}' or '{gtd}'. defaults to 'gtc'
    #     :returns: A dictionary with the following keys:

    #         - type: 'OPTION'
    #         - id: ID of order
    #         - symbol: symbol of asset

    #     :raises Exception: There is an error in the order process.
    #     """
    #     if quantity is None:
    #         quantity = self.get_asset_max_quantity(symbol)
    #     return self.trader.buy_option(symbol, quantity, in_force)

    # def sell_option(
    #     self, symbol: str = None, quantity: int = None, in_force: str = "gtc"
    # ):
    #     """Sells the specified option.

    #     When called, a limit sell order is placed with a limit
    #     price 5% lower than the current price.

    #     If the option symbol is specified, it will sell that option. If it is not, then the
    #     method will select the first stock symbol in the watchlist, and sell all options
    #     related to that stock.

    #     :param str? symbol: Symbol of the asset to sell, in {OCC} format. defaults to sell all options for the first stock in watchlist
    #     :param float? quantity:  Quantity of asset to sell. defaults to sells all
    #     :param str? in_force:  Duration the order is in force. '{gtc}' or '{gtd}'. defaults to 'gtc'
    #     :returns: A dictionary with the following keys:

    #         - type: 'OPTION'
    #         - id: ID of order
    #         - symbol: symbol of asset

    #     :raises Exception: There is an error in the order process.
    #     """
    #     if symbol is None:
    #         symbol = self.watchlist[0]
    #         symbols = [
    #             s["symbol"]
    #             for s in self.get_account_option_positions()
    #             if s["base_symbol"] == symbol
    #         ]
    #     else:
    #         symbols = [symbol]
    #     for s in symbols:
    #         debugger.debug(f"Algo SELL OPTION: {s}")
    #         if quantity is None:
    #             quantity = self.get_asset_quantity(s)
    #         return self.trader.sell_option(s, quantity, in_force)

    def filter_option_chain(
        self,
        symbol=None,
        type=None,
        lower_exp=None,
        upper_exp=None,
        lower_strike=None,
        upper_strike=None,
    ):
        """
        Automatically buys an option that satisfies the criteria specified.

        :param str? symbol: Symbol of stock. defaults to first symbol in watchlist
        :param str? type: 'call' or 'put'
        :param lower_exp: Minimum expiration date of the option.
        :param upper_exp: Maximum expiration date of the option.
        :param float lower_strike: The minimum strike price of the option
        :param float upper_strike: The maximum strike price of the option

        """
        if symbol is None:
            symbol = self.watchlist[0]
        lower_exp = convert_input_to_datetime(lower_exp, self.trader.timezone)
        upper_exp = convert_input_to_datetime(upper_exp, self.trader.timezone)

        exp_dates = self.get_option_chain_info(symbol)["exp_dates"]
        if lower_exp is not None:
            exp_dates = list(filter(lambda x: x >= lower_exp, exp_dates))
        if upper_exp is not None:
            exp_dates = list(filter(lambda x: x <= upper_exp, exp_dates))
        exp_dates = sorted(exp_dates)

        exp_date = exp_dates[0]
        print(f"Exp date: {exp_date}\n")
        chain = self.get_option_chain(symbol, exp_date)
        if lower_strike is not None:
            chain = chain[chain["strike"] >= lower_strike]
        if upper_strike is not None:
            chain = chain[chain["strike"] <= upper_strike]

        if type is not None:
            chain = chain[chain["type"] == type]

        chain = chain.sort_values(by=["strike", "exp_date"])

        return chain

    # ------------------ Functions to trade options ----------------------

    def get_option_chain_info(self, symbol: str = None):
        """Returns metadata about a stock's option chain

        :param str? symbol: symbol of stock. defaults to first symbol in watchlist
        :returns: A dict with the following keys:
            - exp_dates: List of expiration dates, in the fomrat "YYYY-MM-DD"
            - multiplier: Multiplier of the option, usually 100
        """
        if symbol == None:
            symbol = self.watchlist[0]
        return self.trader.fetch_chain_info(symbol)

    def get_option_chain(self, symbol: str, date):
        """Returns the option chain for the specified symbol and expiration date.

        :param str symbol: symbol of stock
        :param dt.datetime date: date of option expiration
        :returns: A dataframe with the follwing columns:

            - exp_date(datetime.datetime): The expiration date
            - strike(float): Strike price
            - type(str): 'call' or 'put'

        The index is the {OCC} symbol of the option.
        Note that the expiration date is not adjusted to the local time zone.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        date = convert_input_to_datetime(date, self.trader.timezone)
        print(f"Date: {date}\n")
        return self.trader.fetch_chain_data(symbol, date)

    def get_option_market_data(self, symbol: str):
        """Retrieves data of specified option.

        :param str? symbol: {OCC} symbol of option
        :returns: A dictionary:

            - price: price of option
            - ask: ask price
            - bid: bid price

        """
        if symbol is None:
            symbol = self.watchlist[0]
        return self.trader.fetch_option_market_data(symbol)

    # ------------------ Technical Indicators -------------------

    def _default_param(self, symbol, interval, ref, prices):
        if symbol is None:
            symbol = self.watchlist[0]
        if self.trader is None:
            if interval == None:
                interval = Interval.MIN_5
            else:
                interval = interval_string_to_enum(interval)
            if prices is None:
                raise Exception(f"No prices found for symbol {symbol}")
        else:
            if interval is None:
                interval = self.trader.interval[symbol]["interval"]
            else:
                interval = interval_string_to_enum(interval)
            if prices == None:
                prices = self.trader.storage.load(symbol, interval)[symbol][ref]

        return symbol, interval, ref, prices

    def rsi(
        self,
        symbol: str = None,
        period: int = 14,
        interval: Interval = None,
        ref: str = "close",
        prices=None,
    ) -> np.array:
        """Calculate RSI

        :param str? symbol:     Symbol to perform calculation on. defaults to first symbol in watchlist
        :param int? period:     Period of RSI. defaults to 14
        :param str? interval:   Interval to perform the calculation. defaults to interval of algorithm
        :param str? ref:        'close', 'open', 'high', or 'low'. defaults to 'close'
        :param list? prices:    When specified, this function will use the values provided in the
                                list to perform calculations and ignore other parameters. defaults to None
        :returns: A list in numpy format, containing RSI values
        """
        symbol, interval, ref, prices = self._default_param(
            symbol, interval, ref, prices
        )

        if len(prices) < period:
            debugger.warning("Not enough data to calculate RSI, returning None")
            return None

        ohlc = pd.DataFrame(
            {
                "close": np.array(prices),
                "open": np.zeros(len(prices)),
                "high": np.zeros(len(prices)),
                "low": np.zeros(len(prices)),
            }
        )
        return TA.RSI(ohlc, period=period).to_numpy()

    def sma(
        self,
        symbol: str = None,
        period: int = 14,
        interval: Interval = None,
        ref: str = "close",
        prices=None,
    ) -> np.array:
        """Calculate SMA

        :param str? symbol:    Symbol to perform calculation on. defaults to first symbol in watchlist
        :param int? period:    Period of SMA. defaults to 14
        :param str? interval:  Interval to perform the calculation. defaults to interval of algorithm
        :param str? ref:       'close', 'open', 'high', or 'low'. defaults to 'close'
        :param list? prices:    When specified, this function will use the values provided in the
                                list to perform calculations and ignore other parameters. defaults to None
        :returns: A list in numpy format, containing SMA values
        """
        symbol, interval, ref, prices = self._default_param(
            symbol, interval, ref, prices
        )

        if len(prices) < period:
            debugger.warning("Not enough data to calculate SMA, returning None")
            return None

        ohlc = pd.DataFrame(
            {
                "close": np.array(prices),
                "open": np.zeros(len(prices)),
                "high": np.zeros(len(prices)),
                "low": np.zeros(len(prices)),
            }
        )
        return TA.SMA(ohlc, period=period).to_numpy()

    def ema(
        self,
        symbol: str = None,
        period: int = 14,
        interval: Interval = None,
        ref: str = "close",
        prices=None,
    ) -> np.array:
        """Calculate EMA

        :param str? symbol:    Symbol to perform calculation on. defaults to first symbol in watchlist
        :param int? period:    Period of EMA. defaults to 14
        :param str? interval:  Interval to perform the calculation. defaults to interval of algorithm
        :param str? ref:       'close', 'open', 'high', or 'low'. defaults to 'close'
        :param list? prices:    When specified, this function will use the values provided in the
                                list to perform calculations and ignore other parameters. defaults to None
        :returns: A list in numpy format, containing EMA values
        """
        symbol, interval, ref, prices = self._default_param(
            symbol, interval, ref, prices
        )

        if len(prices) < period:
            debugger.warning("Not enough data to calculate EMA, returning None")
            return None

        ohlc = pd.DataFrame(
            {
                "close": np.array(prices),
                "open": np.zeros(len(prices)),
                "high": np.zeros(len(prices)),
                "low": np.zeros(len(prices)),
            }
        )
        return TA.EMA(ohlc, period=period).to_numpy()

    def bbands(
        self,
        symbol: str = None,
        period: int = 14,
        interval: Interval = None,
        ref: str = "close",
        dev: float = 1.0,
        prices=None,
    ) -> Tuple[np.array, np.array, np.array]:
        """Calculate Bollinger Bands

        :param str? symbol:    Symbol to perform calculation on. defaults to first symbol in watchlist
        :param int? period:    Period of BBands. defaults to 14
        :param str? interval:  Interval to perform the calculation. defaults to interval of algorithm
        :param str? ref:       'close', 'open', 'high', or 'low'. defaults to 'close'
        :param float? dev:         Standard deviation of the bands. defaults to 1.0
        :param list? prices:    When specified, this function will use the values provided in the
                                list to perform calculations and ignore other parameters. defaults to None
        :returns: A tuple of numpy lists, each a list of BBand top, average, and bottom values
        """
        symbol, interval, ref, prices = self._default_param(
            symbol, interval, ref, prices
        )

        if len(prices) < period:
            debugger.warning("Not enough data to calculate BBands, returning None")
            return None, None, None

        ohlc = pd.DataFrame(
            {
                "close": np.array(prices),
                "open": np.zeros(len(prices)),
                "high": np.zeros(len(prices)),
                "low": np.zeros(len(prices)),
            }
        )

        t, m, b = TA.BBANDS(
            ohlc, period=period, std_multiplier=dev, MA=TA.SMA(ohlc, period)
        ).T.to_numpy()
        return t, m, b

    def crossover(self, prices_0, prices_1):
        """Performs {crossover analysis} on two sets of price data

        :param list prices_0:  First set of price data.
        :param list prices_1:  Second set of price data
        :returns: 'True' if prices_0 most recently crossed over prices_1, 'False' otherwise

        :raises Exception: If either or both price list has less than 2 values
        """
        if len(prices_0) < 2 or len(prices_1) < 2:
            raise Exception(
                "There must be at least 2 datapoints to calculate crossover"
            )
        return prices_0[-2] < prices_1[-2] and prices_0[-1] > prices_1[-1]

    ############### Getters for Trader properties #################

    def get_asset_quantity(self, symbol: str = None) -> float:
        """Returns the quantity owned of a specified asset.

        Assets that are currently pending to be sold are not counted.

        :param str? symbol:  Symbol of asset. defaults to first symbol in watchlist
        :returns: Quantity of asset as float. 0 if quantity is not owned.
        :raises:
        """
        if symbol is None:
            symbol = self.watchlist[0]

        return self.trader.get_asset_quantity(symbol, exclude_pending_sell=True)

    def get_asset_cost(self, symbol: str = None) -> float:
        """Returns the average cost of a specified asset.

        :param str? symbol:  Symbol of asset. defaults to first symbol in watchlist
        :returns: Average cost of asset. Returns None if asset is not being tracked.
        :raises Exception: If symbol is not currently owned.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        if len(symbol) <= 6:
            search = self.trader.stock_positions + self.trader.crypto_positions
            for p in search:
                if p["symbol"] == symbol:
                    return p["avg_price"]
        else:
            for p in self.trader.option_positions:
                if p["occ_symbol"].replace(" ", "") == symbol.replace(" ", ""):
                    return p["avg_price"]

        raise Exception(f"{symbol} is not currently owned")

    def get_asset_price(self, symbol: str = None) -> float:
        """Returns the current price of a specified asset.

        :param str? symbol: Symbol of asset. defaults to first symbol in watchlist
        :returns:           Price of asset.
        :raises Exception:  If symbol is not in the watchlist.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        if len(symbol) <= 6:
            return self.trader.storage.load(symbol, self.interval)[symbol]["close"][-1]
        for p in self.trader.option_positions:
            if p["occ_symbol"] == symbol:
                return p["current_price"] * p["multiplier"]
        return self.get_option_market_data(symbol)["price"] * 100

    def get_asset_price_list(
        self, symbol: str = None, interval: str = None, ref: str = "close"
    ):
        """Returns a list of recent prices for an asset.

        This function is not compatible with options.

        :param str? symbol:     Symbol of stock or crypto asset. defaults to first symbol in watchlist
        :param str? interval:   Interval of data. defaults to the interval of the algorithm
        :param str? ref:        'close', 'open', 'high', or 'low'. defaults to 'close'
        :returns: List of prices
        """
        if symbol is None:
            symbol = self.watchlist[0]
        if interval is None:
            interval = self.interval
        else:
            interval = interval_string_to_enum(interval)
        if len(symbol) <= 6:
            return list(self.trader.storage.load(symbol, interval)[symbol][ref])
        debugger.warning("Price list not available for options")
        return None

    def get_asset_candle(self, symbol: str, interval=None) -> pd.DataFrame():
        """Returns the most recent candle as a pandas DataFrame

        This function is not compatible with options.

        :param str? symbol:  Symbol of stock or crypto asset. defaults to first symbol in watchlist
        :returns: Price of asset as a dataframe with the following columns:

            - open
            - high
            - low
            - close
            - volume

        The index is a datetime object

        :raises Exception: If symbol is not in the watchlist.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        if interval is None:
            interval = self.interval
        if len(symbol) <= 6:
            df = self.trader.storage.load(symbol, interval).iloc[[-1]][symbol]
            print(self.trader.storage.load(symbol, interval))
            return pandas_timestamp_to_local(df, self.trader.timezone)
        debugger.warning("Candles not available for options")
        return None

    def get_asset_candle_list(
        self, symbol: str = None, interval=None
    ) -> pd.DataFrame():
        """Returns the candles of an asset as a pandas DataFrame

        This function is not compatible with options.

        :param str? symbol:  Symbol of stock or crypto asset. defaults to first symbol in watchlist
        :returns: Prices of asset as a dataframe with the following columns:

            - open
            - high
            - low
            - close
            - volume

        The index is a datetime object

        :raises Exception: If symbol is not in the watchlist.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        if interval is None:
            interval = self.interval
        df = self.trader.storage.load(symbol, interval)[symbol]
        return pandas_timestamp_to_local(df, self.trader.timezone)

    def get_asset_returns(self, symbol=None) -> float:
        """Returns the return of a specified asset.

        :param str? symbol:  Symbol of stock, crypto, or option. Options should be in {OCC} format.
                        defaults to first symbol in watchlist
        :returns: Return of asset, expressed as a decimal.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        cost = self.get_asset_cost(symbol)
        # For options, apply the multiplier
        if len(symbol) > 6:
            cost = cost * 100
        price = self.get_asset_price(symbol)
        return (price - cost) / cost

    def get_asset_max_quantity(self, symbol=None):
        """Calculates the maximum quantity of an asset that can be bought given the current buying power.

        :param str? symbol:  Symbol of stock, crypto, or option. Options should be in {OCC} format.
            defaults to first symbol in watchlist
        :returns: Quantity that can be bought.
        """
        if symbol is None:
            symbol = self.watchlist[0]

        power = self.get_account_buying_power()
        price = self.get_asset_price(symbol)
        debugger.debug(f"{symbol} price: {price}, buying power: {power}")
        if is_crypto(symbol):
            price = mark_up(price)
            return math.floor(power / price * 10 ** 5) / 10 ** 5
        else:
            price = mark_up(price)
            return math.floor(power / price)

    def get_account_buying_power(self) -> float:
        """Returns the current buying power of the user

        :returns: The current buying power as a float.
        """
        return self.trader.account["buying_power"]

    def get_account_equity(self) -> float:
        """Returns the current equity.

        :returns: The current equity as a float.
        """
        return self.trader.account["equity"]

    def get_account_stock_positions(self) -> List:
        """Returns the current stock positions.

        :returns: A list of dictionaries with the following keys:
            - symbol
            - quantity
            - avg_price
        """
        return self.trader.stock_positions

    def get_account_crypto_positions(self) -> List:
        """Returns the current crypto positions.

        :returns: A list of dictionaries with the following keys:
            - symbol
            - quantity
            - avg_price
        """
        return self.trader.crypto_positions

    def get_account_option_positions(self) -> List:
        """Returns the current option positions.

        :returns: A list of dictionaries with the following keys:
            - symbol
            - quantity
            - avg_price
        """
        return [
            {
                "symbol": p["symbol"],
                "base_symbol": p["base_symbol"],
                "quantity": p["quantity"],
                "avg_price": p["avg_price"],
            }
            for p in self.trader.option_positions
        ]

    def get_watchlist(self) -> List:
        """Returns the current watchlist."""
        return self.watchlist

    def get_stock_watchlist(self) -> List:
        """Returns the current watchlist."""
        return [s for s in self.watchlist if not is_crypto(s)]

    def get_crypto_watchlist(self) -> List:
        """Returns the current watchlist."""
        return [s for s in self.watchlist if is_crypto(s)]

    def get_time(self):
        """Returns the current hour and minute.

        This returns the current time, which is different from the timestamp
        on a ticker. For example, if you are running an algorithm every 5 minutes,
        at 11:30am you will get a ticker for 11:25am. This function will return
        11:30am.

        :returns: The current time as a datetime object
        """
        return self.get_datetime().time()

    def get_date(self):
        """Returns the current date.

        :returns: The current date as a datetime object
        """
        return self.get_datetime().date()

    def get_datetime(self):
        """Returns the current date and time.

        This returns the current time, which is different from the timestamp
        on a ticker. For example, if you are running an algorithm every 5 minutes,
        at 11:30am you will get a ticker for 11:25am. This function will return
        11:30am.

        :returns: The current date and time as a datetime object
        """
        return datetime_utc_to_local(
            self.trader.streamer.timestamp, self.trader.timezone
        )

    def get_option_position_quantity(self, symbol: str = None) -> bool:
        """Returns the number of types of options held for a stock.

        :param str symbol:  Symbol of the stock to check
        :returns: True if the user has an option for the specified symbol.
        """
        if symbol is None:
            symbol = self.watchlist[0]
        pos = [p for p in self.trader.option_positions if p["symbol"] == symbol]
        return len(pos)

    def is_day_trade(self, action, symbol=None) -> bool:
        """
        Checks if performing a buy or sell will be considered day trading
        """
        if symbol is None:
            symbol = self.watchlist[0]

        history = self.trader.logger.get_transactions()
        if len(history) < 1:
            return False
        recent = history.loc[self.get_date() :]
        recent = recent[recent["symbol"] == symbol]
        if action == "buy":
            return False
        if action == "sell":
            recent = recent.loc[recent["action"] == "buy"]
            return len(recent.index) > 0

    # Used for testing
    def add_symbol(self, symbol: str):
        """Adds a symbol to the watchlist.

        :param str symbol: Symbol of stock or crypto asset.
        """
        self.watchlist.append(symbol)



# Builtins
import datetime as dt
from typing import Any, Dict, List, Tuple
import os.path

# External libraries
import pandas as pd
from webull import webull, paper_webull
import yaml

# Submodule imports
from harvest.api._base import API
from harvest.utils import *


class Webull(API):
    def __init__(self, path: str = None, paper_trader: bool = False):
        super().__init__(path)
        self.paper = paper_trader
        self.wb_tokens = None
        self.timestamp = now()
        wb_filename = "webull_credentials.json"
        if os.path.isfile(wb_filename):
            self.wb_tokens = pd.read_pickle(wb_filename)
        self.api = webull()
        if self.paper:
            debugger.debug("Using Webull Paper Account...")
            print("Using Webull Paper Account...")
            self.api = paper_webull()
        if self.wb_tokens or hasattr(self, "config"):
            self.login()

    def login(self):
        debugger.debug("Logging into Webull...")
        wb_tokens = self.wb_tokens
        if wb_tokens:
            debugger.debug("Trying token login")
            try:
                ret = self.api.api_login(
                    access_token=wb_tokens["accessToken"],
                    refresh_token=wb_tokens["refreshToken"],
                    token_expire=wb_tokens["tokenExpireTime"],
                    uuid=wb_tokens["uuid"],
                )
                if not self.api.is_logged_in():
                    debugger.debug(f"Token login failed. \n{e}")
                    wb_tokens = None
            except Exception as e:
                debugger.debug(f"Token login failed. \n{e}")
                wb_tokens = None
        if not wb_tokens and hasattr(self, "config"):
            debugger.debug("Trying interactive login.")
            self.api.login(
                self.config["wb_username"], self.config["wb_password"], save_token=True
            )
        debugger.debug(f"Logged-in?: {self.api.is_logged_in()}, Paper: {self.paper}")

    def refresh_cred(self):
        super().refresh_cred()
        # self.api.logout()
        self.api.refresh_login(save_token=True)
        # self.login()

    def enter_live_trade_pin(self):
        if not hasattr(self, "config"):
            return
        if self.paper:
            return
        return self.api.get_trade_token(self.config["wb_trade_pin"])

    def setup(self, interval: Dict, trader_main=None):
        super().setup(interval, trader_main)
        self.watch_stock = []
        self.watch_crypto = []
        self.watch_crypto_fmt = []

        val, unit = expand_interval(self.poll_interval)
        if unit == "MIN":
            self.interval_fmt = f"m{val}"
        elif unit == "HR":
            self.interval_fmt = f"h{val}"

        for s in interval:
            if is_crypto(s):
                self.watch_crypto_fmt.append(s[1:])
                self.watch_crypto.append(s)
            else:
                self.watch_stock.append(s)

        self.__option_cache = {}

    def exit(self):
        pass

    def main(self):
        df_dict = {}
        df_dict.update(self.fetch_latest_stock_price())
        df_dict.update(self.fetch_latest_crypto_price())

        self.trader_main(df_dict)

    @API._exception_handler
    def fetch_latest_stock_price(self):
        df = {}
        for s in self.watch_stock:
            ret = self.api.get_bars(stock=s, interval=self.interval_fmt)
            if len(ret) == 0:
                continue
            df_tmp = pd.DataFrame.from_dict(ret)
            df_tmp = self._format_df(df_tmp, [s], self.interval).iloc[[-1]]
            df[s] = df_tmp

        return df

    @API._exception_handler
    def fetch_latest_crypto_price(self):
        df = {}
        for s in self.watch_crypto_fmt:
            ret = self.api.get_bars(stock=s, interval=self.interval_fmt)
            df_tmp = pd.DataFrame.from_dict(ret)
            df_tmp = self._format_df(df_tmp, ["@" + s], self.interval).iloc[[-1]]
            df["@" + s] = df_tmp

        return df

    # -------------- Streamer methods -------------- #

    @API._exception_handler
    def fetch_price_history(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ):

        if start is None:
            start = dt.datetime(1970, 1, 1)
        if end is None:
            end = dt.datetime.now()

        df = pd.DataFrame()

        if start >= end:
            return df

        val, unit = expand_interval(interval)
        if unit == "MIN":
            self.interval_fmt = f"m{val}"
        elif unit == "HR":
            self.interval_fmt = f"h{val}"

        delta = end - start
        delta = delta.total_seconds()
        delta = delta / 3600
        period = 1
        timeframe = 1
        if interval == "DAY" and delta < 24:
            return df
        if delta < 1 or interval == "15SEC" or interval == "1MIN":
            span = "hour"
            period = 1
            timeframe = 1
        elif delta >= 1 and delta < 24 or interval in ["5MIN", "15MIN", "30MIN", "1HR"]:
            span = "day"
        elif delta >= 24 and delta < 24 * 28:
            span = "month"
        elif delta < 24 * 300:
            span = "year"
        else:
            span = "5year"

        if is_crypto(symbol):
            df = self.api.get_bars(
                symbol[1:],
                interval=self.interval_fmt,
                count=int((390 * int(period)) / int(timeframe)),
            )
        else:
            df = self.api.get_bars(
                symbol,
                interval=self.interval_fmt,
                count=int((390 * int(period)) / int(timeframe)),
            )

        df = self._format_df(df, [symbol], interval)

        return df

    @API._exception_handler
    def fetch_chain_info(self, symbol: str):
        ret = self.api.get_options_expiration_dates(symbol)
        return {
            "id": "n/a",
            "exp_dates": [str_to_date(s["date"]) for s in ret],
            "multiplier": 100,
        }

    @API._exception_handler
    def fetch_chain_data(self, symbol: str, date: dt.datetime):

        if (
            bool(self.__option_cache)
            and symbol in self.__option_cache
            and date in self.__option_cache[symbol]
        ):
            return self.__option_cache[symbol][date]

        ret = self.api.get_options(stock=symbol, expireDate=date_to_str(date))
        exp_date = []
        strike = []
        typ = []
        id_ = []
        occ = []
        for entry in ret:
            date = entry["call"]["expireDate"]
            date = str_to_date(date)
            price = float(entry["strikePrice"])
            if entry.get("call"):
                exp_date.append(date)
                strike.append(price)
                typ.append("call")
                id_.append(entry["call"]["tickerId"])
                occ.append(self.data_to_occ(symbol, date, "call", price))
            if entry.get("put"):
                exp_date.append(date)
                strike.append(price)
                typ.append("put")
                id_.append(entry["put"]["tickerId"])
                occ.append(self.data_to_occ(symbol, date, "put", price))

        df = pd.DataFrame(
            {
                "occ_symbol": occ,
                "exp_date": exp_date,
                "strike": strike,
                "type": typ,
                "id": id_,
            }
        )

        df = df.set_index("occ_symbol")
        if not symbol in self.__option_cache:
            self.__option_cache[symbol] = {}
        self.__option_cache[symbol][date] = df

        return df

    @API._exception_handler
    def fetch_option_market_data(self, symbol: str):
        sym, date, _, price = self.occ_to_data(symbol)
        date = str_to_date(date_to_str(date))

        oc_id = self.__option_cache[sym][date][
            self.__option_cache[sym][date].index == symbol
        ].id[0]

        ret = self.api.get_option_quote(stock=sym, optionId=oc_id)
        if not ret.get("data"):
            debugger.error(f"Error in fetch_option_market_data.\nReturned: {ret}")
            raise Exception(f"Error in fetch_option_market_data.\nReturned: {ret}")
        try:
            price = float(ret["data"][0]["close"])
        except:
            price = (
                float(ret["data"][0]["askList"][0]["price"])
                + float(ret["data"][0]["bidList"][0]["price"])
            ) / 2

        return {
            "price": price,
            "ask": float(ret["data"][0]["askList"][0]["price"]),
            "bid": float(ret["data"][0]["bidList"][0]["price"]),
        }

    # ------------- Broker methods ------------- #

    @API._exception_handler
    def fetch_stock_positions(self):
        ret = self.api.get_positions()
        pos = []
        for r in ret:
            if r.get("assetType") and r.get("assetType") != "stock":
                continue
            pos.append(
                {
                    "symbol": r["ticker"]["disSymbol"],
                    "avg_price": float(r["costPrice"]),
                    "quantity": float(r["position"]),
                }
            )
        return pos

    @API._exception_handler
    def fetch_option_positions(self):
        ret = self.api.get_positions()
        pos = []
        for r in ret:
            if not r.get("assetType") or r.get("assetType") != "OPTION":
                continue
            # Get option data such as expiration date
            data = self.api.get_option_quote(
                stock=r["ticker"]["tickerId"], optionId=r["tickerId"]
            )
            pos.append(
                {
                    "base_symbol": r["ticker"]["symbol"],
                    "avg_price": float(r["cost"])
                    / float(data["data"][0]["quoteMultiplier"]),
                    "quantity": float(r["position"]),
                    "multiplier": float(data["data"][0]["quoteMultiplier"]),
                    "exp_date": data["data"][0]["expireDate"],
                    "strike_price": float(data["data"][0]["strikePrice"]),
                    "type": data["data"][0]["direction"],
                }
            )
            pos[-1]["symbol"] = data["data"][0]["symbol"]

        return pos

    @API._exception_handler
    def fetch_crypto_positions(self, key=None):
        ret = self.api.get_positions()
        pos = []
        for r in ret:
            if not r.get("assetType") or r.get("assetType") != "crypto":
                continue
            qty = float(r["position"])

            pos.append(
                {
                    "symbol": "@" + r["ticker"]["symbol"][:-3],
                    "avg_price": float(r["cost"]) / qty,
                    "quantity": qty,
                }
            )
        return pos

    @API._exception_handler
    def update_option_positions(self, positions: List[Any]):
        for r in positions:
            sym, date, _, price = self.occ_to_data(r["occ_symbol"])
            oc_id = self.__option_cache[sym][date][
                self.__option_cache[sym][date].index == symbol
            ].id[0]
            ret = self.api.get_option_quote(stock=sym, optionId=oc_id)

            r["current_price"] = float(ret["data"][0]["close"])
            r["market_value"] = float(ret["data"][0]["close"]) * r["quantity"]
            r["cost_basis"] = r["avg_price"] * r["quantity"]

    def fmt_fetch_account(self, val, data):
        for line in data:
            if line["key"] == val:
                return line["value"]
        return -1

    @API._exception_handler
    def fetch_account(self):
        if not self.api.is_logged_in():
            return None
        ret = self.api.get_account()["accountMembers"]
        return {
            "equity": float(self.fmt_fetch_account("totalMarketValue", ret)),
            "cash": float(self.fmt_fetch_account("cashBalance", ret))
            if not self.paper
            else float(self.fmt_fetch_account("usableCash", ret)),
            "buying_power": float(self.fmt_fetch_account("dayBuyingPower", ret))
            if not self.paper
            else float(self.fmt_fetch_account("usableCash", ret)),
            "bp_options": float(self.fmt_fetch_account("optionBuyingPower", ret)),
            "bp_crypto": float(self.fmt_fetch_account("cryptoBuyingPower", ret)),
            "multiplier": float(-1),
        }

    @API._exception_handler
    def fetch_stock_order_status(self, id):
        ret = self.api.get_history_orders(status="All")
        for r in ret:
            if r["orders"][0]["orderId"] == id:
                return {
                    "type": "STOCK",
                    "id": r["orders"][0]["orderId"],
                    "symbol": r["orders"][0]["ticker"]["symbol"],
                    "price": r.get("lmtPrice"),
                    "avg_price": r.get("avgFilledPrice"),
                    "quantity": r.get("totalQuantity"),
                    "filled_quantity": r.get("filledQuantity"),
                    "side": r["action"],
                    "time_in_force": r["timeInForce"],
                    "status": r["status"].lower(),
                }

    @API._exception_handler
    def fetch_option_order_status(self, id):
        ret = self.api.get_history_orders(status="All")
        for r in ret:
            if r["orders"][0]["orderId"] == id:
                return {
                    "type": "OPTION",
                    "id": r["orders"][0]["orderId"],
                    "symbol": self.data_to_occ(
                        r["orders"][0]["symbol"],
                        str_to_date(r["orders"][0]["optionExpireDate"]),
                        r["orders"][0]["optionType"],
                        float(r["orders"][0]["optionExercisePrice"]),
                    ),
                    "price": r.get("lmtPrice"),
                    "avg_price": r.get("avgFilledPrice"),
                    "qty": r["quantity"],
                    "filled_qty": r["filledQuantity"],
                    "side": r["orders"][0]["optionType"],
                    "time_in_force": r["timeInForce"],
                    "status": r["status"].lower(),
                }

    @API._exception_handler
    def fetch_crypto_order_status(self, id):
        ret = self.api.get_history_orders(status="All")
        for r in ret:
            if r["orderId"] == id:
                return {
                    "type": "CRYPTO",
                    "id": r["orders"][0]["orderId"],
                    "symbol": f"@{r['orders'][0]['ticker']['symbol'].replace('USD', '')}",
                    "qty": float(r["quantity"]),
                    "filled_qty": float(r["cumulative_quantity"]),
                    "filled_price": float(r["executions"][0]["effective_price"])
                    if len(r["executions"])
                    else 0,
                    "filled_cost": float(r["rounded_executed_notional"]),
                    "side": r["side"],
                    "time_in_force": r["timeInForce"],
                    "status": r["status"].lower(),
                }

    @API._exception_handler
    def fetch_order_queue(self):
        queue = []
        ret = self.api.get_current_orders()
        for r in ret:
            queue.append(
                {
                    "type": "STOCK" if self.paper else r["assetType"].upper(),
                    "id": r["orderId"],
                    "symbol": r["ticker"]["symbol"],
                    "price": r.get("lmtPrice"),
                    # " avg_price": r.get["orders"][0].get("avgFilledPrice"),
                    "quantity": r["totalQuantity"],
                    "filled_qty": r["filledQuantity"],
                    "time_in_force": r["timeInForce"],
                    "status": r["status"].lower(),
                    "side": r["action"],
                }
            )
        return queue

    # Order functions are not wrapped in the exception handler to prevent duplicate
    # orders from being made.
    def order_stock_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        ret = None
        if not self.enter_live_trade_pin():
            debugger.error("Error while setting trade pin.")
            raise Exception("Error while setting trade pin.")

        try:
            ret = self.api.place_order(
                stock=symbol,
                tId=None,
                price=limit_price,
                action=side.upper(),
                orderType="LMT",
                enforce=in_force.upper(),
                quant=quantity,
                outsideRegularTradingHour=extended,
            )
            typ = "STOCK"
            if not ret.get("success"):
                debugger.error(f"Error while placing order.\nReturned: {ret}")
                raise Exception("Error while placing order.")
            return {"type": typ, "id": ret["data"]["orderId"], "symbol": symbol}
        except Exception as e:
            debugger.error(
                f"Error while placing order.\nReturned: {ret}", exc_info=True
            )
            raise Exception("Error while placing order.")

    def order_crypto_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        ret = None
        if not self.enter_live_trade_pin():
            debugger.error("Error while setting trade pin.")
            raise Exception("Error while setting trade pin.")

        try:
            ret = self.api.place_order_crypto(
                stock=symbol,
                tId=None,
                price=limit_price,
                action=side.upper(),
                orderType="LMT",
                enforce=in_force.upper(),
                entrust_type="QTY",
                quant=quantity,
                outsideRegularTradingHour=extended,
            )
            typ = "CRYPTO"
            if not ret.get("success"):
                debugger.error(f"Error while placing order.\nReturned: {ret}")
                raise Exception("Error while placing order.")
            return {"type": typ, "id": ret["data"]["orderId"], "symbol": symbol}
        except Exception as e:
            debugger.error(
                f"Error while placing order.\nReturned: {ret}", exc_info=True
            )
            raise Exception("Error while placing order.")

    def order_option_limit(
        self,
        side: str,
        symbol: str,
        quantity: int,
        limit_price: float,
        option_type,
        exp_date: dt.datetime,
        strike,
        in_force: str = "gtc",
    ):
        self.enter_live_trade_pin()
        ret = None
        sym = self.data_to_occ(symbol, exp_date, side, strike)
        date = str_to_date(date_to_str(exp_date))
        oc_id = (
            self.__option_cache[symbol][date][
                self.__option_cache[symbol][date].index == sym
            ]
            .id[0]
            .item()
        )

        if not isinstance(oc_id, int):
            debugger.error(
                "Error while placing order_option_limit. Can't find optionId."
            )
            raise Exception("Error while placing order.")
        try:
            ret = self.api.place_order_option(
                optionId=oc_id,
                lmtPrice=limit_price,
                stpPrice=None,
                action=side.upper(),
                orderType="LMT",
                enforce=in_force.upper(),
                quant=quantity,
            )

            if not ret or not ret.get("orderId"):
                debugger.error(f"Error while placing order.\nReturned: {ret}")
                raise Exception("Error while placing order.")
            return {"type": "OPTION", "id": ret["orderId"], "symbol": symbol}

        except:
            debugger.error(
                f"Error while placing order.\nReturned: {ret}", exc_info=True
            )
            raise Exception("Error while placing order")

    def _format_df(
        self, df: pd.DataFrame, watch: List[str], interval: str, latest: bool = False
    ):
        df = df[["open", "close", "high", "low", "volume"]].astype(float)
        df.columns = pd.MultiIndex.from_product([watch, df.columns])
        return df.dropna()

    def create_secret(self, path):
        import harvest.wizard as wizard

        w = wizard.Wizard()

        w.println(
            "Hmm, looks like you haven't set up login credentials for Webull yet."
        )
        should_setup = w.get_bool("Do you want to set it up now?", default="y")

        if not should_setup:
            w.println("Webull has limited functionality when not logged in.")
            w.println("You can set up the credentials manually, or use other brokers.")
            return False

        w.println("Alright! Let's get started")

        have_account = w.get_bool("Do you have a Webull account?", default="y")
        if not have_account:
            w.println(
                "In that case you'll first need to make an account. I'll wait here, so hit Enter or Return when you've done that."
            )
            w.wait_for_input()

        username = w.get_string("Username: ")
        password = w.get_password("Password: ")
        device = w.get_string("Device Name (e.g. Harvester): ")
        w.println(f"Getting security question.")
        from webull import webull

        wb = webull()
        sec_question = wb.get_security(username)[0]

        if not sec_question:
            w.println("Come back after setting up a security question.")
            return False

        sec_answer = w.get_string(f"{sec_question['questionName']}: ")

        if sec_question["questionId"] == "1001":
            sec_check = len(sec_answer)
            while sec_check != 4:
                new_passcode = w.get_bool(
                    "The code should be only 4 digits... do you want to re-try?",
                    default="n",
                )
                if new_passcode:
                    sec_answer = w.get_string(f"{sec_question['questionName']}: ")
                    sec_check = len(sec_answer)
                else:
                    break

        pin = w.get_password("Enter trade PIN for live trading: ")

        while len(pin) != 6:
            new_pin = w.get_bool(
                "The pin code should be only 6 digits... do you want to re-try?",
                default="n",
            )
            if new_pin:
                pin = w.get_password("PIN for live trading: ")
            else:
                break

        w.println(f"Requesting MFA code via email.")
        wb.get_mfa(username)
        mfa = w.get_password("MFA Code: ")

        ret = wb.login(
            username,
            password,
            device,
            mfa,
            sec_question["questionId"],
            sec_answer,
            save_token=True,
        )

        if not wb.is_logged_in():
            w.println(f"Login failed... check your info and try again.\nReason: {ret}")
            return False

        if isinstance(pin, int) and not wb.get_trade_token(pin):
            w.println(
                f"Trade PIN verification failed... check your info and try again.\nReason: {ret}"
            )
            return False

        wb.logout()
        w.println(f"All steps are complete now 🎉. Generating secret.yml...")

        d = {"wb_username": username, "wb_password": password, "wb_trade_pin": pin}

        with open(path, "w") as file:
            yml = yaml.dump(d, file)

        w.println(
            f"secret.yml has been created! Make sure you keep this file somewhere secure and never share it with other people."
        )

        return True



# Builtins
import pathlib
import unittest
import datetime as dt
import os

from harvest.utils import not_gh_action


class TestWebull(unittest.TestCase):
    @not_gh_action
    def test_fetch_prices(self):
        from harvest.api.webull import Webull

        wb = Webull()
        interval = {
            "@BTC": {"interval": Interval.MIN_1, "aggregations": []},
            "SPY": {"interval": Interval.MIN_1, "aggregations": []},
        }
        wb.setup(interval)
        df = wb.fetch_price_history("SPY", interval=Interval.MIN_1)["SPY"]
        self.assertEqual(
            sorted(list(df.columns.values)),
            sorted(["open", "high", "low", "close", "volume"]),
        )
        df = wb.fetch_price_history("BTC", interval=Interval.MIN_1)["BTC"]
        self.assertEqual(
            sorted(list(df.columns.values)),
            sorted(["open", "high", "low", "close", "volume"]),
        )

    @not_gh_action
    def test_setup(self):
        from harvest.api.webull import Webull

        wb = Webull()
        watch = ["SPY", "@BTC"]
        interval = {
            "@BTC": {"interval": Interval.MIN_1, "aggregations": []},
            "SPY": {"interval": Interval.MIN_1, "aggregations": []},
        }
        wb.setup(interval)
        self.assertEqual(wb.watch_stock, ["SPY"])
        self.assertEqual(wb.watch_crypto, ["@BTC"])

    @not_gh_action
    def test_main(self):
        from harvest.api.webull import Webull

        def test_main(df):
            self.assertEqual(len(df), 2)
            self.assertEqual(df["SPY"].columns[0][0], "SPY")
            self.assertEqual(df["@BTC"].columns[0][0], "@BTC")

        wb = Webull()
        watch = ["SPY", "@BTC"]
        interval = {
            "@BTC": {"interval": Interval.MIN_1, "aggregations": []},
            "SPY": {"interval": Interval.MIN_1, "aggregations": []},
        }
        wb.setup(interval, test_main)
        wb.main()

    @not_gh_action
    def test_main_single(self):
        from harvest.api.webull import Webull

        def test_main(df):
            self.assertEqual(len(df), 1)
            self.assertEqual(df["SPY"].columns[0][0], "SPY")

        wb = Webull()
        watch = ["SPY"]
        interval = {"SPY": {"interval": Interval.MIN_1, "aggregations": []}}
        wb.setup(interval, test_main)
        wb.main()

    @not_gh_action
    def test_chain_info(self):
        from harvest.api.webull import Webull

        wb = Webull()
        watch = ["SPY"]
        interval = {"SPY": {"interval": Interval.MIN_1, "aggregations": []}}
        wb.setup(interval)
        info = wb.fetch_chain_info("SPY")
        self.assertGreater(len(info["exp_dates"]), 0)

    @not_gh_action
    def test_chain_data(self):
        from harvest.api.webull import Webull

        wb = Webull()
        watch = ["LMND"]
        interval = {"LMND": {"interval": Interval.MIN_1, "aggregations": []}}
        wb.setup(interval)
        dates = wb.fetch_chain_info("LMND")["exp_dates"]
        data = wb.fetch_chain_data("LMND", dates[0])
        self.assertGreater(len(data), 0)
        self.assertListEqual(list(data.columns), ["exp_date", "strike", "type", "id"])
        sym = data.index[0]
        df = wb.fetch_option_market_data(sym)
        self.assertTrue(True)

    @not_gh_action
    def test_order_option_limit(self):
        dummy = PaperBroker()
        dummy.streamer = Webull()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        exp_date = dt.datetime.now() + dt.timedelta(hours=5)
        order = dummy.order_option_limit(
            "buy", "A", 5, 50000, "OPTION", exp_date, 50001
        )
        self.assertEqual(order["type"], "OPTION")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

    @not_gh_action
    def test_sell(self):
        dummy = PaperBroker()
        dummy.streamer = Webull()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.sell("A", 2)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

    @not_gh_action
    def test_sell_order_limit(self):
        dummy = PaperBroker()
        dummy.streamer = Webull()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.order_limit("sell", "A", 2, 50000)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

    @not_gh_action
    def test_buy(self):
        dummy = PaperBroker()
        dummy.streamer = Webull()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.buy("A", 5)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")

    @not_gh_action
    def test_buy_order_limit(self):
        dummy = PaperBroker()
        dummy.streamer = Webull()
        interval = {"A": {"interval": Interval.MIN_1, "aggregations": []}}
        dummy.setup(interval)
        order = dummy.order_limit("buy", "A", 5, 50000)
        self.assertEqual(order["type"], "STOCK")
        self.assertEqual(order["id"], 0)
        self.assertEqual(order["symbol"], "A")


if __name__ == "__main__":
    unittest.main()



import subprocess
from setuptools import Command, setup


class LintCMD(Command):
    user_options = [
        (
            "check=",
            "c",
            "if True check if files pass linting checks; don't actually lint",
        )
    ]

    def initialize_options(self):
        self.check = False

    def finalize_options(self):
        self.check = self.check == "True"

    def run(self):
        command = ["black", "harvest"]

        if self.check:
            command.append("--check")

        exit(subprocess.run(command).returncode)


class CoverageTestCMD(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        a = subprocess.run(
            ["coverage", "run", "-m", "unittest", "discover", "-s", "tests"]
        ).returncode
        b = subprocess.run(["coverage", "report"]).returncode
        c = subprocess.run(["coverage", "html"]).returncode
        exit(a + b + c)


setup(
    cmdclass={
        "lint": LintCMD,
        "test": CoverageTestCMD,
    },
    include_package_data=True,
)



import re
from os import listdir, makedirs
from os.path import isfile, join
import pandas as pd
import datetime as dt
from typing import Tuple

from harvest.storage import BaseStorage
from harvest.utils import *


"""
This module serves as a storage system for pandas dataframes in with csv files.
"""


class CSVStorage(BaseStorage):
    """
    An extension of the basic storage that saves data in csv files.
    """

    def __init__(self, save_dir: str = "data"):
        super().__init__()
        """
        Adds a directory to save data to. Loads any data that is currently in the
        directory.
        """
        self.save_dir = save_dir

        # if the data dir does not exists, create it
        makedirs(self.save_dir, exist_ok=True)

        files = [f for f in listdir(self.save_dir) if isfile(join(self.save_dir, f))]

        for file in files:
            debugger.debug(file)
            file_search = re.search("^(@?[\w]+)@([\w]+).csv$", file)
            symbol, interval = file_search.group(1), file_search.group(2)
            interval = interval_string_to_enum(interval)
            data = pd.read_csv(join(self.save_dir, file), index_col=0, parse_dates=True)
            data.index = pd.to_datetime(data.index, unit="s")
            data.columns = pd.MultiIndex.from_product([[symbol], data.columns])
            super().store(symbol, interval, data)

    def store(
        self,
        symbol: str,
        interval: Interval,
        data: pd.DataFrame,
        remove_duplicate: bool = True,
    ) -> None:
        """
        Stores the stock data in the storage dictionary as a csv file.
        :symbol: a stock or crypto
        :interval: the interval between each data point, must be atleast
             1 minute
        :data: a pandas dataframe that has stock data and has a datetime
            index
        """
        super().store(symbol, interval, data, remove_duplicate)

        if not data.empty:
            self.storage_lock.acquire()
            self.storage_price[symbol][interval][symbol].to_csv(
                self.save_dir + f"/{symbol}@{interval_enum_to_string(interval)}.csv"
            )
            self.storage_lock.release()



# Builtins
import unittest
import datetime as dt
import os

from harvest.utils import *
from harvest.api.polygon import PolygonStreamer


class TestPolygonStreamer(unittest.TestCase):
    @not_gh_action
    def test_fetch_prices(self):
        poly = PolygonStreamer("poly_secret.yaml", True)
        df = poly.fetch_price_history(
            "AAPL", Interval.HR_1, now() - dt.timedelta(days=7), now()
        )["AAPL"]
        self.assertEqual(
            list(df.columns.values), ["open", "high", "low", "close", "volume"]
        )


if __name__ == "__main__":
    unittest.main()



# Builtins
import datetime as dt
from typing import Any, Dict, List, Tuple
import hashlib

# External libraries
import pytz
import pandas as pd
import numpy as np

# Submodule imports
from harvest.api._base import API
from harvest.utils import *


class DummyStreamer(API):
    """DummyStreamer, as its name implies, is a dummy broker class that can
    be useful for testing algorithms. When used as a streamer, it will return
    randomly generated prices.
    """

    interval_list = [
        Interval.MIN_1,
        Interval.MIN_5,
        Interval.MIN_15,
        Interval.MIN_30,
        Interval.HR_1,
        Interval.DAY_1,
    ]

    default_timestamp = dt.datetime(year=2000, month=1, day=1, hour=0, minute=0)

    def __init__(
        self,
        timestamp: dt.datetime = default_timestamp,
        realistic_times: bool = False,
    ):

        super().__init__(None)

        self.trader_main = None
        self.realistic_times = realistic_times

        # Set the current time
        self._set_timestamp(timestamp)
        # Used so `fetch_price_history` can work without running `setup`
        self.interval = self.interval_list[0]
        # Store random values and generates for each asset tot make `fetch_price_history` fixed
        self.randomness = {}

    def start(self) -> None:
        pass

    def main(self):
        df_dict = {}
        df_dict.update(self.fetch_latest_stock_price())
        df_dict.update(self.fetch_latest_crypto_price())

        self.trader_main(df_dict)

    def fetch_latest_stock_price(self) -> Dict[str, pd.DataFrame]:
        """
        Gets fake stock data in the last three day interval and returns the last
        value. The reason the last three days are needed is because no data is returned
        when the stock market is closed, e.g. weekends.
        """

        results = {}
        today = self.timestamp
        last = today - dt.timedelta(days=3)

        for symbol in self.interval:
            if not is_crypto(symbol):
                results[symbol] = self.fetch_price_history(
                    symbol, self.interval[symbol]["interval"], last, today
                ).iloc[[-1]]
        return results

    def fetch_latest_crypto_price(self) -> Dict[str, pd.DataFrame]:
        """
        Gets fake crypto data in the last three day interval and  returns the last
        value. The reason the last three days are needed is because no data is returned
        when the stock market is closed, e.g. weekends.
        """

        results = {}
        today = self.timestamp
        last = today - dt.timedelta(days=3)
        for symbol in self.interval:
            if is_crypto(symbol):
                results[symbol] = self.fetch_price_history(
                    symbol, self.interval[symbol]["interval"], last, today
                ).iloc[[-1]]
        return results

    # -------------- Streamer methods -------------- #

    def get_current_time(self):
        return self.timestamp

    def fetch_price_history(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ) -> pd.DataFrame:

        if start is None:
            if interval in [
                Interval.MIN_1,
                Interval.MIN_5,
                Interval.MIN_15,
                Interval.MIN_30,
            ]:
                start = self.timestamp - dt.timedelta(days=2)
            elif interval == Interval.HR_1:
                start = self.timestamp - dt.timedelta(days=14)
            else:
                start = self.timestamp - dt.timedelta(days=365)

        if end is None:
            end = self.timestamp

        if start.tzinfo is None or start.tzinfo.utcoffset(start) is None:
            start = pytz.utc.localize(start)

        if end.tzinfo is None or end.tzinfo.utcoffset(end) is None:
            end = pytz.utc.localize(end)

        # Convert datetime to indices
        start_index = int((start - epoch_zero()).total_seconds() // 60)
        end_index = 1 + int((end - epoch_zero()).total_seconds() // 60)

        if symbol in self.randomness:
            # If we already generated data from this asset

            num_of_random = 1 + end_index

            if len(self.randomness[symbol]) < num_of_random:
                # If the new end index is greater than the data we have

                # Get the rng for this symbol
                rng = self.randomness[symbol + "_rng"]

                # Generate a bunch of random numbers from -0.5 to 0.5
                increments = (
                    rng.random(num_of_random - len(self.randomness[symbol])) - 0.499
                )

                # Get sum of the previous price changes
                latest_price_change = self.randomness[symbol][-1]

                # Calculate the change in price since the first price
                new_price_changes = latest_price_change + np.cumsum(increments)

                # Store the new prices
                self.randomness[symbol] = np.append(
                    self.randomness[symbol], new_price_changes
                )

        else:
            # If there is no information about the asset

            # Create an rng using the asset's symbol as a seed
            rng = np.random.default_rng(int.from_bytes(symbol.encode("ascii"), "big"))
            num_of_random = 1 + end_index
            increments = rng.random(num_of_random) - 0.499

            # Store the price change since the first price
            self.randomness[symbol] = np.cumsum(increments)
            self.randomness[symbol + "_rng"] = rng

        # The inital price is arbitarly calculated from the first change in price
        start_price = 1000 * (self.randomness[symbol][0] + 0.501)

        times = []
        current_time = start

        # Get the prices for the current interval
        prices = start_price + self.randomness[symbol][start_index:end_index]
        # Prevent prices from going negative
        prices[prices < 0] = 0.01

        # Calculate olhcv from the prices
        open_s = 0.95 * prices
        low = 0.8 * prices
        high = 1.2 * prices
        close = 1.05 * prices
        volume = (1000 * (prices + 20)).astype(int)

        # Fake the timestamps
        while current_time <= end:
            times.append(current_time)
            current_time += dt.timedelta(minutes=1)

        d = {
            "timestamp": times,
            "open": open_s,
            "high": high,
            "low": low,
            "close": close,
            "volume": volume,
        }

        results = pd.DataFrame(data=d).set_index("timestamp")

        if self.realistic_times:
            debugger.debug(
                "Dummy Broker excluding information when the stock market is closed."
            )
            open_time = dt.time(hour=13, minute=30)
            close_time = dt.time(hour=20)

            # Removes datapoints when the stock marked is closed. Does not handle holidays.
            results = results.loc[
                (open_time < results.index.time) & (results.index.time < close_time)
            ]
            results = results[
                (results.index.dayofweek != 5) & (results.index.dayofweek != 6)
            ]

        results.columns = pd.MultiIndex.from_product([[symbol], results.columns])
        results = aggregate_df(results, interval)
        return results

    # TODO: Generate dummy option data

    def fetch_option_market_data(self, symbol: str):
        # This is a placeholder so Trader doesn't crash
        message = hashlib.sha256()
        message.update(symbol.encode("utf-8"))
        message.update(str(self.timestamp).encode("utf-8"))
        hsh = message.digest()
        price = int.from_bytes(hsh[:4], "big") / (2 ** 32)
        price = (price + 1) * 1.5
        debugger.debug(
            f"Dummy Streamer fake fetch option market data price for {symbol}: {price}"
        )

        return {
            "price": price,
            "ask": price * 1.05,
            "bid": price * 0.95,
        }

    # Not implemented functions:
    #   fetch_chain_info
    #   fetch_chain_data

    # ------------- Broker methods ------------- #

    # Not implemented functions:
    #   fetch_stock_positions
    #   fetch_option_positions
    #   fetch_crypto_positions
    #   update_option_positions
    #   fetch_account
    #   fetch_stock_order_status
    #   fetch_option_order_status
    #   fetch_crypto_order_status
    #   fetch_order_queue

    # --------------- Methods for Trading --------------- #

    # Not implemented functions:
    #   order_stock_limit
    #   order_crypto_limit
    #   order_option_limit

    # ------------- Helper methods ------------- #

    def _set_timestamp(self, current_datetime: dt.datetime) -> None:
        if (
            current_datetime.tzinfo is None
            or current_datetime.tzinfo.utcoffset(current_datetime) is None
        ):
            self.timestamp = pytz.utc.localize(current_datetime)
        else:
            self.timestamp = current_datetime

    def tick(self) -> None:
        self.timestamp += interval_to_timedelta(self.poll_interval)
        if not self.trader_main == None:
            self.main()



# Builtins
import datetime as dt
from typing import Any, Dict, List, Tuple

# External libraries
import tzlocal
import pandas as pd
import yfinance as yf

# Submodule imports
from harvest.api._base import API
from harvest.utils import *


class YahooStreamer(API):

    interval_list = [
        Interval.MIN_1,
        Interval.MIN_5,
        Interval.MIN_15,
        Interval.MIN_30,
        Interval.HR_1,
    ]
    exchange = "NASDAQ"

    def __init__(self, path=None):
        self.timestamp = now()

    def setup(self, interval: Dict, trader_main=None):
        super().setup(interval, trader_main)

        self.watch_ticker = {}

        for s in interval:
            if is_crypto(s):
                self.watch_ticker[s] = yf.Ticker(s[1:] + "-USD")
            else:
                self.watch_ticker[s] = yf.Ticker(s)

        self.option_cache = {}

    def fmt_interval(self, interval):
        val, unit = expand_interval(interval)
        if unit == "MIN":
            interval_fmt = f"{val}m"
        elif unit == "HR":
            interval_fmt = f"{val}h"

        return interval_fmt

    def fmt_symbol(self, symbol: str):
        return symbol[1:] + "-USD" if is_crypto(symbol) else symbol

    def unfmt_symbol(self, symbol: str):
        if symbol[-4:] == "-USD":
            return "@" + symbol[:-4]
        return symbol

    def exit(self):
        self.option_cache = {}

    def main(self):
        df_dict = {}
        combo = [
            self.fmt_symbol(sym)
            for sym in self.interval
            if is_freq(self.timestamp, self.interval[sym]["interval"])
        ]

        if len(combo) == 1:
            s = combo[0]
            interval_fmt = self.fmt_interval(
                self.interval[self.unfmt_symbol(s)]["interval"]
            )
            df = yf.download(s, period="1d", interval=interval_fmt, prepost=True)
            debugger.debug(f"From yfinance got: {df}")
            if len(df.index) == 0:
                return
            s = self.unfmt_symbol(s)
            df = self._format_df(df, s)
            df_dict[s] = df
        else:
            names = " ".join(combo)
            df = None
            required_intervals = {}
            for s in combo:
                i = self.interval[self.unfmt_symbol(s)]["interval"]
                if i in required_intervals:
                    required_intervals[i].append(s)
                else:
                    required_intervals[i] = [s]
            for i in required_intervals:
                names = " ".join(required_intervals[i])
                df_tmp = yf.download(
                    names, period="1d", interval=self.fmt_interval(i), prepost=True
                )
                debugger.debug(f"From yfinance got: {df_tmp}")
                df = df_tmp if df is None else df.join(df_tmp)

            if len(df.index) == 0:
                return
            for s in combo:
                df_tmp = df.iloc[:, df.columns.get_level_values(1) == s]
                if len(df_tmp.index) == 0:
                    continue
                df_tmp.columns = df_tmp.columns.droplevel(1)
                if s[-4:] == "-USD":
                    s = "@" + s[:-4]
                df_tmp = self._format_df(df_tmp, s)
                df_dict[s] = df_tmp

        debugger.debug(f"From yfinance dict: {df_dict}")
        self.trader_main(df_dict)

    # -------------- Streamer methods -------------- #

    @API._exception_handler
    def fetch_price_history(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ):

        debugger.debug(f"Fetching {symbol} {interval} price history")
        if isinstance(interval, str):
            interval = interval_string_to_enum(interval)

        if start is None:
            start = epoch_zero()
        if end is None:
            end = now()

        df = pd.DataFrame()

        if start >= end:
            return df

        val, unit = expand_interval(interval)
        if unit == "MIN":
            get_fmt = f"{val}m"
        elif unit == "HR":
            get_fmt = f"{val}h"
        else:
            get_fmt = "1d"

        if interval == Interval.MIN_1:
            period = "5d"
        elif interval >= Interval.MIN_5 and interval <= Interval.HR_1:
            period = "1mo"
        else:
            period = "max"

        crypto = False
        if is_crypto(symbol):
            symbol = symbol[1:] + "-USD"
            crypto = True

        df = yf.download(symbol, period=period, interval=get_fmt, prepost=True)
        if crypto:
            symbol = "@" + symbol[:-4]
        df = self._format_df(df, symbol)
        df = df.loc[start:end]
        return df

    @API._exception_handler
    def fetch_chain_info(self, symbol: str):
        option_list = self.watch_ticker[symbol].options
        return {
            "id": "n/a",
            "exp_dates": [
                convert_input_to_datetime(s, tzlocal.get_localzone())
                for s in option_list
            ],
            "multiplier": 100,
        }

    @API._exception_handler
    def fetch_chain_data(self, symbol: str, date: dt.datetime):

        if (
            bool(self.option_cache)
            and symbol in self.option_cache
            and date in self.option_cache[symbol]
        ):
            return self.option_cache[symbol][date]

        df = pd.DataFrame(columns=["contractSymbol", "exp_date", "strike", "type"])

        chain = self.watch_ticker[symbol].option_chain(date_to_str(date))
        puts = chain.puts
        puts["type"] = "put"
        calls = chain.calls
        calls["type"] = "call"
        df = df.append(puts)
        df = df.append(calls)

        df = df.rename(columns={"contractSymbol": "occ_symbol"})
        df["exp_date"] = df.apply(
            lambda x: self.occ_to_data(x["occ_symbol"])[1], axis=1
        )
        df = df[["occ_symbol", "exp_date", "strike", "type"]]
        df.set_index("occ_symbol", inplace=True)

        if symbol not in self.option_cache:
            self.option_cache[symbol] = {}
        self.option_cache[symbol][date] = df

        return df

    @API._exception_handler
    def fetch_option_market_data(self, occ_symbol: str):
        occ_symbol = occ_symbol.replace(" ", "")
        symbol, date, typ, _ = self.occ_to_data(occ_symbol)
        chain = self.watch_ticker[symbol].option_chain(date_to_str(date))
        chain = chain.calls if typ == "call" else chain.puts
        df = chain[chain["contractSymbol"] == occ_symbol]
        debugger.debug(occ_symbol, df)
        return {
            "price": float(df["lastPrice"].iloc[0]),
            "ask": float(df["ask"].iloc[0]),
            "bid": float(df["bid"].iloc[0]),
        }

    # ------------- Broker methods ------------- #

    # Not implemented functions:
    #   fetch_stock_positions
    #   fetch_option_positions
    #   fetch_crypto_positions
    #   update_option_positions
    #   fetch_account
    #   fetch_stock_order_status
    #   fetch_option_order_status
    #   fetch_crypto_order_status
    #   fetch_order_queue

    # --------------- Methods for Trading --------------- #

    # Not implemented functions:
    #   order_limit
    #   order_option_limit

    # ------------- Helper methods ------------- #

    def _format_df(self, df: pd.DataFrame, symbol: str):
        df = df.copy()
        df.reset_index(inplace=True)
        ts_name = df.columns[0]
        df["timestamp"] = df[ts_name]
        df = df.set_index(["timestamp"])
        d = df.index[0]
        if d.tzinfo is None or d.tzinfo.utcoffset(d) is None:
            df = df.tz_localize("UTC")
        else:
            df = df.tz_convert(tz="UTC")
        df = df.drop([ts_name], axis=1)
        df = df.rename(
            columns={
                "Open": "open",
                "Close": "close",
                "High": "high",
                "Low": "low",
                "Volume": "volume",
            }
        )
        df = df[["open", "high", "low", "close", "volume"]].astype(float)

        df.columns = pd.MultiIndex.from_product([[symbol], df.columns])

        return df.dropna()



# Builtins
import os
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from harvest.storage.database_storage import DBStorage
from harvest.utils import gen_data


class TestDBStorage(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.db_file = "foo.db"

        # Create the database file
        with open(self.db_file, "w") as f:
            pass

    def test_simple_load(self):
        storage = DBStorage("sqlite:///foo.db")
        data = gen_data("A", 50)
        storage.store("A", "1MIN", data.copy(True))

        loaded_data = storage.load("A", "1MIN")

        # sort_index to prevent order issues
        assert_frame_equal(loaded_data.sort_index(axis=1), data.sort_index(axis=1))

    def test_reset(self):
        storage = DBStorage("sqlite:///foo.db")
        data = gen_data("A", 50)
        storage.store("A", "1MIN", data.copy(True))
        storage.reset("A", "1MIN")
        loaded_data = storage.load("A", "1MIN")

        self.assertTrue(loaded_data is None)

    @classmethod
    def tearDownClass(self):
        if os.path.exists(self.db_file):
            os.remove(self.db_file)


if __name__ == "__main__":
    unittest.main()



# Builtins
import yaml
import datetime as dt
from typing import Any, Dict, List, Tuple

# External libraries
import krakenex
import pandas as pd

# Submodule imports
from harvest.api._base import API
from harvest.utils import *


class Kraken(API):

    interval_list = [Interval.MIN_1, Interval.MIN_5, Interval.HR_1, Interval.DAY_1]

    crypto_ticker_to_kraken_names = {
        "BTC": "XXBTZ",
        "ETH": "XETH",
        "ADA": "ADA",
        "USDT": "USDT",
        "XRP": "XXRP",
        "SOL": "SOL",
        "DOGE": "XDG",
        "DOT": "DOT",
        "USDC": "USDC",
        "UNI": "UNI",
        "LTC": "XLTC",
        "LINK": "LINK",
        "BCH": "BCH",
        "FIL": "FIL",
        "MATIC": "MATIC",
        "WBTC": "WBTC",
        "ETC": "XETC",
        "XLM": "XXLM",
        "TRX": "TRX",
        "DAI": "DAI",
        "EOS": "EOS",
        "ATOM": "ATOM",
        "AAVE": "AAVE",
        "XMR": "XXMR",
        "AXS": "AXS",
        "GRT": "GRT",
        "XTZ": "XXTZ",
        "ALGO": "ALGO",
        "MKR": "MKR",
        "KSM": "KSM",
        "WAVE": "WAVE",
        "COMP": "COMP",
        "DASH": "DASH",
        "CHZ": "CHZ",
        "ZEC": "XZEC",
        "MANA": "MANA",
        "ENJ": "ENJ",
        "SUSHI": "SUSHI",
        "YFI": "YFI",
        "QTUM": "QTUM",
        "FLOW": "FLOW",
        "SNX": "SNX",
        "BAT": "BAT",
        "SC": "SC",
        "ICX": "ICX",
        "PERP": "PERP",
        "BNT": "BNT",
        "OMG": "OMG",
        "CRV": "CRV",
        "ZRX": "ZRX",
        "NANO": "NANO",
        "ANKR": "ANKR",
        "SAND": "SAND",
        "REN": "REN",
        "KAVA": "KAVA",
        "MINA": "MINA",
        "1INCH": "1INCH",
        "GHST": "GHST",
        "ANT": "ANT",
        "REP": "XREP",
        "REPV2": "XREPV2",
        "BADGER": "BADGER",
        "BAL": "BAL",
        "BAND": "BAND",
        "CTSI": "CTSI",
        "CQT": "CQT",
        "EWT": "EWT",
        "MLN": "XMLN",
        "ETH2": "ETH2",
        "GNO": "GNO",
        "INJ": "INJ",
        "KAR": "KAR",
        "KEEP": "KEEP",
        "KNC": "KNC",
        "LSK": "LSK",
        "LTP": "LTP",
        "LRC": "LRC",
        "MIR": "MIR",
        "OCEAN": "OCEAN",
        "PAXG": "PAXG",
        "RARI": "RARI",
        "REN": "REN",
        "XRP": "XXRP",
        "SRM": "SRM",
        "STORJ": "STORJ",
        "TBTC": "TBTC",
        "OGN": "OGN",
        "OXT": "OXT",
    }

    kraken_names_to_crypto_ticker = {
        v: k for k, v in crypto_ticker_to_kraken_names.items()
    }

    def __init__(self, path: str = None):
        super().__init__(path)
        self.api = krakenex.API(self.config["api_key"], self.config["secret_key"])

    def setup(self, interval: Dict, trader_main=None):
        super().setup(interval, trader_main)
        self.watch_crypto = []
        for sym in interval:
            if is_crypto(sym):
                self.watch_crypto.append(sym)
            else:
                debugger.warning(f"Kraken does not support stocks. Ignoring {sym}.")

        self.option_cache = {}

    def exit(self):
        self.option_cache = {}

    def main(self):
        df_dict = {}
        df_dict.update(self.fetch_latest_crypto_price())

        self.trader_main(df_dict)

    @API._exception_handler
    def fetch_latest_crypto_price(self):
        dfs = {}
        for symbol in self.watch_crypto:
            dfs[symbol] = self.fetch_price_history(
                symbol,
                self.interval[symbol]["interval"],
                now() - dt.timedelta(days=7),
                now(),
            ).iloc[[-1]]
        return dfs

    # -------------- Streamer methods -------------- #

    @API._exception_handler
    def fetch_price_history(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ):

        debugger.debug(f"Fetching {symbol} {interval} price history")

        if start is None:
            start = now() - dt.timedelta(days=365 * 5)
        elif not has_timezone(start):
            start = set_system_timezone(start)
        if end is None:
            end = now()
        elif not has_timezone(end):
            end = set_system_timezone(end)

        if start >= end:
            return pd.DataFrame()

        if interval not in self.interval_list:
            raise Exception(
                f"Interval {interval} not in interval list. Possible options are: {self.interval_list}"
            )
        val, unit = expand_interval(interval)
        df = self.get_data_from_kraken(symbol, val, unit, start, end)

        return df

    def fetch_chain_info(self, symbol: str):
        raise NotImplementedError("Kraken does not support options.")

    def fetch_chain_data(self, symbol: str, date: dt.datetime):
        raise NotImplementedError("Kraken does not support options.")

    def fetch_option_market_data(self, occ_symbol: str):
        raise NotImplementedError("Kraken does not support options.")

    # ------------- Broker methods ------------- #

    @API._exception_handler
    def fetch_stock_positions(self):
        debugger.error("Kraken does not support stocks. Returning an empty list.")
        return []

    @API._exception_handler
    def fetch_option_positions(self):
        debugger.error("Kraken does not support options")
        return []

    @API._exception_handler
    def fetch_crypto_positions(self):
        positions = self.get_result(self.api.query_private("OpenPositions"))

        def fmt(crypto: Dict[str, Any]):
            # Remove the currency
            symbol = crypto["pair"][:-4]
            # Convert from kraken name to crypto currency ticker
            symbol = kraken_name_to_crypto_ticker.get(symbol)
            return {
                "symbol": "@" + symbol,
                "avg_price": float(crypto["cost"]) / float(crypto["vol"]),
                "quantity": float(crypto["vol"]),
                "kraken": crypto,
            }

        return [fmt(pos) for pos in positions]

    def update_option_positions(self, positions: List[Any]):
        debugger.error("Kraken does not support options. Doing nothing.")

    @API._exception_handler
    def fetch_account(self):
        account = self.get_result(self.api.query_private("Balance"))
        if account is None:
            equity = 0
            cash = 0
        else:
            equity = sum(float(v) for k, v in account.items() if k != "ZUSD")
            cash = account.get("ZUSD", 0)
        return {
            "equity": equity,
            "cash": cash,
            "buying_power": equity + cash,
            "multiplier": 1,
            "kraken": account,
        }

    def fetch_stock_order_status(self, order_id: str):
        return NotImplementedError("Kraken does not support stocks.")

    def fetch_option_order_status(self, order_id: str):
        raise Exception("Kraken does not support options.")

    @API._exception_handler
    def fetch_crypto_order_status(self, order_id: str):
        order = self.api.query_private("QueryOrders", {"txid": order_id})
        symbol = kraken_names_to_crypto_ticker.get(crypto["descr"]["pair"][:-4])
        return {
            "type": "CRYPTO",
            "symbol": "@" + symbol,
            "id": crypto.key(),
            "quantity": float(crypto["vol"]),
            "filled_quantity": float(crypto["vol_exec"]),
            "side": crypto["descr"]["type"],
            "time_in_force": None,
            "status": crypto["status"],
            "kraken": crypto,
        }

    # --------------- Methods for Trading --------------- #

    @API._exception_handler
    def fetch_order_queue(self):
        open_orders = self.get_result(self.api.query_private("OpenOrders"))
        open_orders = open_orders["open"]

        def fmt(crypto: Dict[str, Any]):
            symbol = kraken_names_to_crypto_ticker.get(crypto["descr"]["pair"][:-4])
            return {
                "type": "CRYPTO",
                "symbol": "@" + symbol,
                "id": crypto.key(),
                "quantity": float(crypto["vol"]),
                "filled_quantity": float(crypto["vol_exec"]),
                "side": crypto["descr"]["type"],
                "time_in_force": None,
                "status": crypto["status"],
                "kraken": crypto,
            }

        return [fmt(order) for order in open_orders]

    def order_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        if is_crypto(symbol):
            symbol = ticker_to_kraken(symbol)
        else:
            raise Exception("Kraken does not support stocks.")

        order = self.get_result(
            self.api.query_private(
                "AddOrder",
                {
                    "ordertype": "limit",
                    "type": side,
                    "volume": quantity,
                    "pair": symbol,
                },
            )
        )

        return {
            "type": "CRYPTO",
            "id": order["txid"],
            "symbol": symbol,
            "kraken": order,
        }

    def order_option_limit(
        self,
        side: str,
        symbol: str,
        quantity: int,
        limit_price: float,
        option_type,
        exp_date: dt.datetime,
        strike,
        in_force: str = "gtc",
    ):
        raise NotImplementedError("Kraken does not support options.")

    # ------------- Helper methods ------------- #

    def get_data_from_kraken(
        self,
        symbol: str,
        multiplier: int,
        timespan: str,
        start: dt.datetime,
        end: dt.datetime,
    ) -> pd.DataFrame:
        if timespan == "MIN":
            multiplier *= 1
        elif timespan == "HR":
            multiplier *= 60
        elif timespan == "DAY":
            multiplier *= 1440

        if is_crypto(symbol):
            temp_symbol = self.ticker_to_kraken(symbol)
        else:
            raise Exception("Kraken does not support stocks.")
        bars = self.get_result(
            self.api.query_public(
                "OHLC",
                {"pair": temp_symbol, "interval": multiplier, "since": end.timestamp},
            )
        )
        df = pd.DataFrame(
            bars[temp_symbol],
            columns=[
                "timestamp",
                "open",
                "high",
                "low",
                "close",
                "vwap",
                "volume",
                "count",
            ],
        )
        df = self._format_df(df, symbol)
        df = df.loc[start:end]
        return df

    def _format_df(self, df: pd.DataFrame, symbol: str):
        df = df[["timestamp", "open", "high", "low", "close", "volume"]].astype(float)
        df.index = pd.to_datetime(df["timestamp"], unit="s", utc=True)
        df = df.drop(columns=["timestamp"])
        df.columns = pd.MultiIndex.from_product([[symbol], df.columns])

        return df.dropna()

    def ticker_to_kraken(self, ticker: str):
        if not is_crypto(ticker):
            raise Exception("Kraken does not support stocks.")

        if ticker[1:] in self.crypto_ticker_to_kraken_names:
            # Currently Harvest supports trades for USD and not other currencies.
            kraken_ticker = self.crypto_ticker_to_kraken_names.get(ticker[1:]) + "USD"
            asset_pairs = self.get_result(self.api.query_public("AssetPairs")).keys()
            if kraken_ticker in asset_pairs:
                return kraken_ticker
            else:
                raise Exception(f"{kraken_ticker} is not a valid asset pair.")
        else:
            raise Exception(f"Kraken does not support ticker {ticker}.")

    def get_result(self, response: Dict[str, Any]):
        """Given a kraken response from an endpoint, either raise an error if an
        error exists or return the data in the results key.
        """
        if len(response["error"]) > 0:
            raise Exception("\n".join(response["error"]))
        return response.get("result", None)

    def create_secret(self, path: str) -> bool:
        import harvest.wizard as wizard

        w = wizard.Wizard()

        w.println("Hmm, looks like you haven't set up an api key for Kraken.")
        should_setup = w.get_bool("Do you want to set it up now?", default="y")

        if not should_setup:
            w.println("You can't use Kraken without an API key.")
            w.println(
                "You can set up the credentials manually, or use other streamers."
            )
            return False

        w.println("Alright! Let's get started")

        have_account = w.get_bool("Do you have an Kraken account?", default="y")
        if not have_account:
            w.println(
                "In that case you'll first need to make an account. This takes a few steps."
            )
            w.println(
                "First visit: https://www.kraken.com/sign-up and sign up. Hit Enter or Return for the next step."
            )
            w.wait_for_input()
            w.println(
                "Create an account, go to your account dropdown > Security > API and create an API key."
            )
            w.wait_for_input()

        api_key_id = w.get_string("Enter your API key ID")
        secret_key = w.get_password("Enter your API secret key")

        w.println(f"All steps are complete now 🎉. Generating {path}...")

        d = {"api_key": f"{api_key_id}", "secret_key": f"{secret_key}"}

        with open(path, "w") as file:
            yml = yaml.dump(d, file)

        w.println(
            f"{path} has been created! Make sure you keep this file somewhere secure and never share it with other people."
        )

        return True



# Builtins
from logging import debug
import re
import threading
import traceback
import sys
from sys import exit
from signal import signal, SIGINT
import time
import datetime as dt

# External libraries
import tzlocal

# Submodule imports
from harvest.utils import *
from harvest.storage import BaseStorage
from harvest.api.yahoo import YahooStreamer
from harvest.api.dummy import DummyStreamer
from harvest.api.paper import PaperBroker
from harvest.storage import BaseStorage
from harvest.server import Server


class LiveTrader:
    """
    :broker: Both the broker and streamer store a Broker object.
        Broker places orders and retrieves latest account info like equity.
    :streamer: Streamer retrieves the latest stock price and calls handler().
    """

    interval_list = [
        Interval.SEC_15,
        Interval.MIN_1,
        Interval.MIN_5,
        Interval.MIN_15,
        Interval.MIN_30,
        Interval.HR_1,
        Interval.DAY_1,
    ]

    def __init__(self, streamer=None, broker=None, storage=None, debug=False):
        """Initializes the Trader."""

        self._init_checks()

        self._set_streamer_broker(streamer, broker)
        # Initialize the storage
        self.storage = BaseStorage() if storage is None else storage

        self._init_attributes()
        self._setup_debugger(debug)

    def _init_checks(self):
        # Harvest only supports Python 3.9 or newer.
        if sys.version_info[0] < 3 or sys.version_info[1] < 9:
            raise Exception("Harvest requires Python 3.9 or above.")

    def _set_streamer_broker(self, streamer, broker):
        """Sets the streamer and broker."""
        # If streamer is not specified, use YahooStreamer
        self.streamer = YahooStreamer() if streamer is None else streamer
        # Broker must be specified
        if broker is None:
            # TODO: Raise exception if specified class is streaming only
            if isinstance(self.streamer, YahooStreamer):
                raise Exception("Broker must be specified")
            else:
                self.broker = self.streamer
        else:
            self.broker = broker

    def _init_attributes(self):

        signal(SIGINT, self.exit)

        self.watchlist_global = []  # List of securities specified in this class
        self.algo = []  # List of algorithms to run.
        self.account = {}  # Local cache of account data.
        self.stock_positions = []  # Local cache of current stock positions.
        self.option_positions = []  # Local cache of current options positions.
        self.crypto_positions = []  # Local cache of current crypto positions.
        self.order_queue = []  # Queue of unfilled orders.

        self.server = Server(self)  # Initialize the web interface server

        self.timezone = tzlocal.get_localzone()
        debugger.debug(f"Timezone: {self.timezone}")

    def _setup_debugger(self, debug):
        # Set up logger
        if debug:
            debugger.setLevel("DEBUG")

        debugger.debug(
            f"Streamer: {type(self.streamer).__name__}\nBroker: {type(self.broker).__name__}\nStorage: {type(self.storage).__name__}"
        )

    def start(
        self,
        interval="5MIN",
        aggregations=[],
        sync=True,
        server=False,
        all_history=True,
    ):
        """Entry point to start the system.

        :param str? interval: The interval to run the algorithm. defaults to '5MIN'
        :param list[str]? aggregations: A list of intervals. The Trader will aggregate data to the intervals specified in this list.
            For example, if this is set to ['5MIN', '30MIN'], and interval is '1MIN', the algorithm will have access to
            5MIN, 30MIN aggregated data in addition to 1MIN data. defaults to None
        :param bool? sync: If true, the system will sync with the broker and fetch current positions and pending orders. defaults to true.
        :param bool? all_history: If true, gets all history for all the given assets and if false only get data in the past three days.
        """
        debugger.debug("Setting up Harvest...")

        # If sync is on, call the broker to load pending orders and all positions currently held.
        if sync:
            self._setup_stats()
            for s in self.stock_positions:
                self.watchlist_global.append(s["symbol"])
            for s in self.option_positions:
                self.watchlist_global.append(s["base_symbol"])
            for s in self.crypto_positions:
                self.watchlist_global.append(s["symbol"])
            for s in self.order_queue:
                sym = s["base_symbol"] if s["type"] == "OPTION" else s["symbol"]
                self.watchlist_global.append(sym)

        # Remove duplicates in watchlist
        self.watchlist_global = list(set(self.watchlist_global))
        debugger.debug(f"Watchlist: {self.watchlist_global}")

        # Initialize a dict of symbols and the intervals they need to run at
        self._setup_params(interval, aggregations)

        if len(self.interval) == 0:
            raise Exception("No securities were added to watchlist")

        # Initialize the account
        self._setup_account()
        self.storage.init_performace_data(
            self.account["equity"], self.streamer.timestamp
        )

        self.broker.setup(self.interval, self.main)
        if self.broker != self.streamer:
            # Only call the streamer setup if it is a different
            # instance than the broker otherwise some brokers can fail!
            self.streamer.setup(self.interval, self.main)

        # Initialize the storage
        self._storage_init(all_history)

        for a in self.algo:
            a.trader = self
            a.setup()

        debugger.debug("Setup complete")

        if server:
            self.server.start()

        self.streamer.start()

    def _setup_stats(self):
        """Initializes local cache of stocks, options, and crypto positions."""

        # Get any pending orders
        ret = self.broker.fetch_order_queue()
        self.order_queue = ret
        debugger.debug(f"Fetched orders:\n{self.order_queue}")

        # Get positions
        pos = self.broker.fetch_stock_positions()
        self.stock_positions = pos
        pos = self.broker.fetch_option_positions()
        self.option_positions = pos
        pos = self.broker.fetch_crypto_positions()
        self.crypto_positions = pos
        debugger.debug(
            f"Fetched positions:\n{self.stock_positions}\n{self.option_positions}\n{self.crypto_positions}"
        )

        # Update option stats
        self.broker.update_option_positions(self.option_positions)
        debugger.debug(f"Updated option positions:\n{self.option_positions}")

    def _setup_params(self, interval, aggregations):
        """
        Sets up configuration parameters for the Trader, notably
        the 'interval' attribute.
        """
        interval = interval_string_to_enum(interval)
        aggregations = [interval_string_to_enum(a) for a in aggregations]
        self.interval = {}

        # Initialize a dict with symbol keys and values indicating
        # what data intervals they need.
        for sym in self.watchlist_global:
            self.interval[sym] = {}
            self.interval[sym]["interval"] = interval
            self.interval[sym]["aggregations"] = aggregations

        # Update the dict based on parameters specified in Algo class
        for a in self.algo:
            a.config()

            # If the algorithm does not specify a parameter, use the one
            # specified in the Trader class
            if len(a.watchlist) == 0:
                a.watchlist = self.watchlist_global
            if a.interval is None:
                a.interval = interval
            else:
                a.interval = interval_string_to_enum(a.interval)
            if a.aggregations is None:
                a.aggregations = aggregations
            else:
                a.aggregations = [interval_string_to_enum(a) for a in a.aggregations]

            # For each symbol specified in the Algo...
            for sym in a.watchlist:
                # If the algorithm needs data for the symbol at a higher frequency than
                # it is currently available in the Trader class, update the interval
                if sym in self.interval:
                    cur_interval = self.interval[sym]["interval"]
                    if a.interval < cur_interval:
                        self.interval[sym]["aggregations"].append(cur_interval)
                        self.interval[sym]["interval"] = a.interval
                # If symbol is not in global watchlist, simply add it
                else:
                    self.interval[sym] = {}
                    self.interval[sym]["interval"] = a.interval
                    self.interval[sym]["aggregations"] = a.aggregations

                # If the algo specifies an aggregation that is currently not set, add it to the
                # global aggregation list
                for agg in a.aggregations:
                    if agg not in self.interval[sym]["aggregations"]:
                        self.interval[sym]["aggregations"].append(agg)

        # Remove any duplicates in the dict
        for sym in self.interval:
            new_agg = list((set(self.interval[sym]["aggregations"])))
            self.interval[sym]["aggregations"] = [] if new_agg is None else new_agg

    def _setup_account(self):
        """Initializes local cache of account info.
        For testing, it should manually be specified
        """
        ret = self.broker.fetch_account()
        if ret is None:
            raise Exception("Failed to load account info from broker.")
        self.account = ret

    def _storage_init(self, all_history: bool):
        """
        Initializes the storage.
        :all_history: bool :
        """

        for sym in self.interval:
            for inter in [self.interval[sym]["interval"]] + self.interval[sym][
                "aggregations"
            ]:
                start = None if all_history else now() - dt.timedelta(days=3)
                df = self.streamer.fetch_price_history(sym, inter, start)
                self.storage.store(sym, inter, df)

    # ================== Functions for main routine =====================

    def main(self, df_dict):
        """
        Main loop of the Trader.
        """
        # Periodically refresh access tokens
        if (
            self.streamer.timestamp.hour % 12 == 0
            and self.streamer.timestamp.minute == 0
        ):
            self.streamer.refresh_cred()

        self.storage.add_performance_data(
            self.account["equity"], self.streamer.timestamp
        )

        # Save the data locally
        for sym in df_dict:
            self.storage.store(sym, self.interval[sym]["interval"], df_dict[sym])

        # Aggregate the data to other intervals
        for sym in df_dict:
            for agg in self.interval[sym]["aggregations"]:
                self.storage.aggregate(sym, self.interval[sym]["interval"], agg)

        # If an order was processed, fetch the latest position info.
        # Otherwise, calculate current positions locally
        update = self._update_order_queue()
        self._update_position_cache(
            df_dict, new=update, option_update=len(self.option_positions) > 0
        )

        new_algo = []
        for a in self.algo:
            if not is_freq(self.streamer.timestamp, a.interval):
                new_algo.append(a)
                continue
            try:
                debugger.info(f"Running algo: {a}")
                a.main()
                new_algo.append(a)
            except Exception as e:
                debugger.warning(
                    f"Algorithm {a} failed, removing from algorithm list.\n"
                )
                debugger.warning(f"Exception: {e}\n")
                debugger.warning(f"Traceback: {traceback.format_exc()}\n")

        if len(new_algo) <= 0:
            debugger.critical("No algorithms to run")
            exit()

        self.algo = new_algo

        self.broker.exit()
        self.streamer.exit()

    def _update_order_queue(self):
        """Check to see if outstanding orders have been accepted or rejected
        and update the order queue accordingly.
        """
        debugger.debug(f"Updating order queue: {self.order_queue}")
        for i, order in enumerate(self.order_queue):
            if "type" not in order:
                raise Exception(f"key error in {order}\nof {self.order_queue}")
            if order["type"] == "STOCK":
                stat = self.broker.fetch_stock_order_status(order["id"])
            elif order["type"] == "OPTION":
                stat = self.broker.fetch_option_order_status(order["id"])
            elif order["type"] == "CRYPTO":
                stat = self.broker.fetch_crypto_order_status(order["id"])
            debugger.debug(f"Updating status of order {order['id']}")
            self.order_queue[i].update(stat)

        debugger.debug(f"Updated order queue: {self.order_queue}")
        new_order = []
        order_filled = False
        for order in self.order_queue:
            # TODO: handle cancelled orders
            if order["status"] == "filled":
                order_filled = True
                debugger.debug(
                    f"Order {order['id']} filled at {order['filled_time']} at {order['filled_price']}"
                )
                self.storage.store_transaction(
                    order["filled_time"],
                    "N/A",
                    order["symbol"],
                    order["side"],
                    order["quantity"],
                    order["filled_price"],
                )
            else:
                new_order.append(order)
        self.order_queue = new_order

        # if an order was processed, update the positions and account info
        return order_filled

    def _update_position_cache(self, df_dict, new=False, option_update=False):
        """Update local cache of stocks, options, and crypto positions"""
        # Update entries in local cache
        # API should also be called if load_watch is false, as there is a high chance
        # that data in local cache are not representative of the entire portfolio,
        # meaning total equity cannot be calculated locally
        if new:
            self._fetch_account_data()
        if option_update:
            self.broker.update_option_positions(self.option_positions)

        debugger.debug(f"Stock positions: {self.stock_positions}")
        debugger.debug(f"Option positions: {self.option_positions}")
        debugger.debug(f"Crypto positions: {self.crypto_positions}")

        if new:
            return

        net_value = 0
        for p in self.stock_positions + self.crypto_positions:
            key = p["symbol"]
            if key in df_dict:
                price = df_dict[key][key]["close"]
            elif key not in self.watchlist_global:
                i = self.streamer.poll_interval
                end = now()
                start = end - interval_to_timedelta(i) * 2
                price = self.streamer.fetch_price_history(key, i, start, end)
                price = price[key]["close"][-1]
            else:
                continue
            p["current_price"] = price
            value = price * p["quantity"]
            p["market_value"] = value
            net_value += value

        equity = net_value + self.account["cash"]
        self.account["equity"] = equity
        self.stock_positions = self.broker.fetch_stock_positions()

    def _fetch_account_data(self):
        pos = self.broker.fetch_stock_positions()
        self.stock_positions = [p for p in pos if p["symbol"] in self.interval]
        pos = self.broker.fetch_option_positions()
        self.option_positions = [p for p in pos if p["base_symbol"] in self.interval]
        pos = self.broker.fetch_crypto_positions()
        self.crypto_positions = [p for p in pos if p["symbol"] in self.interval]
        ret = self.broker.fetch_account()
        self.account = ret

    # --------------------- Interface Functions -----------------------

    def fetch_chain_info(self, *args, **kwargs):
        return self.streamer.fetch_chain_info(*args, **kwargs)

    def fetch_chain_data(self, *args, **kwargs):
        return self.streamer.fetch_chain_data(*args, **kwargs)

    def fetch_option_market_data(self, *args, **kwargs):
        return self.streamer.fetch_option_market_data(*args, **kwargs)

    def buy(self, symbol: str, quantity: int, in_force: str, extended: bool):
        # Check if user has enough buying power
        buy_power = self.account["buying_power"]
        if symbol_type(symbol) == "OPTION":
            price = self.streamer.fetch_option_market_data(symbol)["price"]
        else:
            price = self.storage.load(symbol, self.interval[symbol]["interval"])[
                symbol
            ]["close"][-1]

        limit_price = mark_up(price)
        total_price = limit_price * quantity

        debugger.warning(
            f"Attempting to buy {quantity} shares of {symbol} at price {price} with price limit {limit_price} and a maximum total price of {total_price}"
        )

        if total_price >= buy_power:
            debugger.error(
                "Not enough buying power.\n"
                + f"Total price ({price} * {quantity} * 1.05 = {limit_price*quantity}) exceeds buying power {buy_power}."
                + "Reduce purchase quantity or increase buying power."
            )
            return None

        # TODO? Perform other checks
        ret = self.broker.buy(symbol, quantity, limit_price, in_force, extended)

        if ret is None:
            debugger.debug("BUY failed")
            return None
        self.order_queue.append(ret)
        debugger.debug(f"BUY: {self.streamer.timestamp}, {symbol}, {quantity}")

        return ret

    def sell(self, symbol: str, quantity: int, in_force: str, extended: bool):
        # Check how many of the given asset we currently own
        owned_qty = self.get_asset_quantity(symbol, exclude_pending_sell=True)
        if quantity > owned_qty:
            debugger.debug(
                "SELL failed: More quantities are being sold than currently owned."
            )
            return None

        if symbol_type(symbol) == "OPTION":
            price = self.streamer.fetch_option_market_data(symbol)["price"]
        else:
            price = self.storage.load(symbol, self.interval[symbol]["interval"])[
                symbol
            ]["close"][-1]

        limit_price = mark_down(price)

        ret = self.broker.sell(symbol, quantity, limit_price, in_force, extended)
        if ret is None:
            debugger.debug("SELL failed")
            return None
        self.order_queue.append(ret)
        debugger.debug(f"SELL: {self.streamer.timestamp}, {symbol}, {quantity}")
        return ret

    # ================ Helper Functions ======================
    def get_asset_quantity(
        self, symbol: str = None, include_pending_buy=False, exclude_pending_sell=False
    ) -> float:
        """Returns the quantity owned of a specified asset.

        :param str? symbol:  Symbol of asset. defaults to first symbol in watchlist
        :returns: Quantity of asset as float. 0 if quantity is not owned.
        :raises:
        """
        if symbol is None:
            symbol = self.watchlist_global[0]

        if typ := symbol_type(symbol) == "OPTION":
            owned_qty = sum(
                p["quantity"] for p in self.option_positions if p["symbol"] == symbol
            )
        elif typ == "CRYPTO":
            owned_qty = sum(
                p["quantity"] for p in self.crypto_positions if p["symbol"] == symbol
            )
        else:
            owned_qty = sum(
                p["quantity"] for p in self.stock_positions if p["symbol"] == symbol
            )

        if include_pending_buy:
            owned_qty += sum(
                o["quantity"]
                for o in self.order_queue
                if o["symbol"] == symbol and o["side"] == "buy"
            )

        if exclude_pending_sell:
            owned_qty -= sum(
                o["quantity"]
                for o in self.order_queue
                if o["symbol"] == symbol and o["side"] == "sell"
            )

        return owned_qty

    # def buy_option(self, symbol: str, quantity: int, in_force: str):
    #     ret = self.broker.buy_option(symbol, quantity, in_force)
    #     if ret is None:
    #         raise Exception("BUY failed")
    #     self.order_queue.append(ret)
    #     debugger.debug(f"BUY: {self.streamer.timestamp}, {symbol}, {quantity}")
    #     debugger.debug(f"BUY order queue: {self.order_queue}")
    #     self.logger.add_transaction(self.streamer.timestamp, "buy", "option", symbol, quantity)
    #     return ret

    # def sell_option(self, symbol: str, quantity: int, in_force: str):
    #     owned_qty = sum(
    #         p["quantity"] for p in self.option_positions if p["symbol"] == symbol
    #     )
    #     owned_qty -= sum(
    #         o["quantity"]
    #         for o in self.order_queue
    #         if o["symbol"] == symbol and o["side"] == "sell"
    #     )
    #     if quantity > owned_qty:
    #         debugger.debug("SELL failed: Quantity too high")
    #         return None

    #     ret = self.broker.sell_option(symbol, quantity, in_force)
    #     if ret is None:
    #         raise Exception("SELL failed")
    #     self.order_queue.append(ret)
    #     debugger.debug(f"SELL: {self.streamer.timestamp}, {symbol}, {quantity}")
    #     debugger.debug(f"SELL order queue: {self.order_queue}")
    #     self.logger.add_transaction(self.streamer.timestamp, "sell", "option", symbol, quantity)
    #     return ret

    def set_algo(self, algo):
        """Specifies the algorithm to use.

        :param Algo algo: The algorithm to use. You can either pass in a single Algo class, or a
            list of Algo classes.
        """
        self.algo = algo if isinstance(algo, list) else [algo]

    def add_algo(self, algo):
        self.algo.append(algo)

    def set_symbol(self, symbol):
        """Specifies the symbol(s) to watch.

        Cryptocurrencies should be prepended with an `@` to differentiate them from stocks.
        For example, '@ETH' will refer to Etherium, while 'ETH' will refer to Ethan Allen Interiors.
        If this method was previously called, the symbols specified earlier will be replaced with the
        new symbols.

        :symbol str symbol: Ticker Symbol(s) of stock or cryptocurrency to watch.
            It can either be a string, or a list of strings.
        """
        self.watchlist_global = symbol if isinstance(symbol, list) else [symbol]

    def exit(self, signum, frame):
        # TODO: Gracefully exit
        debugger.debug("\nStopping Harvest...")
        exit(0)


class PaperTrader(LiveTrader):
    """
    A class for trading in the paper trading environment.
    """

    def __init__(self, streamer=None, storage=None, debug=False):
        """Initializes the Trader."""

        self._init_checks()

        # If streamer is not specified, use YahooStreamer
        self.streamer = YahooStreamer() if streamer is None else streamer
        self.broker = PaperBroker(streamer=self.streamer)

        self.storage = (
            BaseStorage() if storage is None else storage
        )  # Initialize the storage
        self._init_attributes()

        self._setup_debugger(debug)



import logging
from flask import Flask, render_template, request, redirect
from werkzeug.security import generate_password_hash, check_password_hash
from flask_cors import CORS
import threading
import json

from flask_login import (
    LoginManager,
    login_user,
    login_required,
    logout_user,
    current_user,
    UserMixin,
)

from harvest.utils import debugger


class User(UserMixin):
    def __init__(self, username, password):
        self.id = username
        self.password = password
        self.default_password = True


class DB:
    def __init__(self):
        self.users = []

    def add_user(self, username, password):
        hashed_password = generate_password_hash(password)
        self.users.append(User(username, hashed_password))

    def update_user_password(self, username, password):
        for user in self.users:
            if user.get_id() == username:
                user.password = generate_password_hash(password)
                user.default_password = False
                return True
        return False

    def set_is_default_password(self, username, is_default):
        for user in self.users:
            if user.get_id() == username:
                user.default_password = is_default

    def get_user(self, username):
        for user in self.users:
            if user.get_id() == username:
                return user
        return None


app = Flask(__name__, template_folder="gui", static_folder="gui", static_url_path="/")

# CORS(app)
login_manager = LoginManager()
login_manager.init_app(app)
login_manager.login_view = "login"

db = DB()  # Initiate database

trader = None


class Server:
    """
    Runs a web server in a new thread.
    """

    def __init__(self, t):
        app.config["SECRET_KEY"] = "secret!"  # TODO: Generate random secret key
        db.add_user("admin", "admin")  # Default user
        global trader
        trader = t

    def start(self):
        debugger.info("Starting web server")
        server = threading.Thread(target=app.run, kwargs={"port": 11111}, daemon=True)
        server.start()


# ========= Backend API endpoints =========


@app.route("/api/login", methods=["POST"])
def api_login():

    username = request.json["username"]
    password = request.json["password"]

    user = db.get_user(username)

    # If user has the right credentials, log the user in.
    if user and check_password_hash(user.password, password):
        login_user(user)
        if user.default_password:
            return redirect("/update_password")
        else:
            return redirect("/")
    else:
        return redirect("/login")


@app.route("/api/update_password", methods=["POST"])
@login_required
def api_update_password():

    username = current_user.get_id()
    new_password = request.json["password"]

    db.update_user_password(username, new_password)
    db.set_is_default_password(username, False)

    logout_user()
    return redirect("/login")


@app.route("/api/logout", methods=["POST"])
@login_required
def api_logout():
    username = current_user.get_id()
    user = db.get_user(username)
    logout_user(user)


@app.route("/api/crypto_positions")
@login_required
def api_crypto_positions():
    return json.dumps(trader.crypto_positions)


@app.route("/api/stock_positions")
@login_required
def api_stock_positions():
    return json.dumps(trader.stock_positions)


@app.route("/api/option_positions")
@login_required
def api_option_positions():
    return json.dumps(trader.option_positions)


# ========= Web GUI endpoints =========


@app.route("/")
@login_required
def interface():
    return render_template("index.html")


@app.route("/login", methods=["GET"])
def login():
    return render_template("minimal/login.html")


@app.route("/update_password", methods=["GET"])
@login_required
def update_password():
    return render_template("minimal/update_password.html")


# ========= Handlers =========


@login_manager.unauthorized_handler
def unauthorized():
    print("Need to log in")
    return json.dumps({"message": "need to log in"})


@login_manager.user_loader
def load_user(user_id):
    print(user_id)
    u = db.get_user(user_id)
    print(f"got {u.is_authenticated}")
    return u



# Builtins
import json
import yaml
import datetime as dt
import urllib.request
from typing import Any, Dict, List, Tuple

# External libraries
import pandas as pd

# Submodule imports
from harvest.api._base import API
from harvest.utils import *


class PolygonStreamer(API):

    interval_list = [Interval.MIN_1, Interval.MIN_5, Interval.HR_1, Interval.DAY_1]

    def __init__(self, path: str = None, is_basic_account: bool = False):
        super().__init__(path)
        self.basic = is_basic_account

    def setup(self, interval, trader_main=None):
        self.watch_stock = []
        self.watch_crypto = []

        for sym in interval:
            if is_crypto(sym):
                self.watch_crypto.append(sym)
            else:
                self.watch_stock.append(sym)

        self.option_cache = {}
        super().setup(interval, trader_main)

    def exit(self):
        self.option_cache = {}

    def main(self):
        df_dict = {}
        combo = self.watch_stock + self.watch_crypto
        if self.basic and len(combo) > 5:
            debugger.error(
                "Basic accounts only allow for 5 API calls per minute, trying to get data for more than 5 assets! Aborting."
            )
            return

        for s in combo:
            df = self.fetch_price_history(
                s, Interval.MIN_1, now() - dt.timedelta(days=1), now()
            ).iloc[-1]
            df_dict[s] = df
            debugger.debug(df)
        self.trader_main(df_dict)

    # -------------- Streamer methods -------------- #

    @API._exception_handler
    def fetch_price_history(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ):

        debugger.debug(f"Fetching {symbol} {interval} price history")

        if start is None:
            start = now() - dt.timedelta(days=365 * 2)
        if end is None:
            end = now()

        if start >= end:
            return pd.DataFrame()

        val, unit = expand_interval(interval)
        return self.get_data_from_polygon(symbol, val, unit, start, end)

    @API._exception_handler
    def fetch_chain_info(self, symbol: str):
        return {
            "id": "n/a",
            "exp_dates": [str_to_date(s) for s in self.watch_ticker[symbol].options],
            "multiplier": 100,
        }

    @API._exception_handler
    def fetch_chain_data(self, symbol: str, date: dt.datetime):

        if (
            bool(self.option_cache)
            and symbol in self.option_cache
            and date in self.option_cache[symbol]
        ):
            return self.option_cache[symbol][date]

        df = pd.DataFrame(columns=["contractSymbol", "exp_date", "strike", "type"])

        chain = self.watch_ticker[symbol].option_chain(date_to_str(date))
        puts = chain.puts
        puts["type"] = "put"
        calls = chain.calls
        calls["type"] = "call"
        df = df.append(puts)
        df = df.append(calls)

        df = df.rename(columns={"contractSymbol": "occ_symbol"})
        df["exp_date"] = df.apply(
            lambda x: self.occ_to_data(x["occ_symbol"])[1], axis=1
        )
        df = df[["occ_symbol", "exp_date", "strike", "type"]]
        df.set_index("occ_symbol", inplace=True)

        if symbol not in self.option_cache:
            self.option_cache[symbol] = {}
        self.option_cache[symbol][date] = df

        return df

    @API._exception_handler
    def fetch_option_market_data(self, occ_symbol: str):
        occ_symbol = occ_symbol.replace(" ", "")
        symbol, date, typ, _ = self.occ_to_data(occ_symbol)
        chain = self.watch_ticker[symbol].option_chain(date_to_str(date))
        chain = chain.calls if typ == "call" else chain.puts
        df = chain[chain["contractSymbol"] == occ_symbol]
        debugger.debug(occ_symbol, df)
        return {
            "price": float(df["lastPrice"].iloc[0]),
            "ask": float(df["ask"].iloc[0]),
            "bid": float(df["bid"].iloc[0]),
        }

    # ------------- Broker methods ------------- #

    # Not implemented functions:
    #   fetch_stock_positions
    #   fetch_option_positions
    #   fetch_crypto_positions
    #   update_option_positions
    #   fetch_account
    #   fetch_stock_order_status
    #   fetch_option_order_status
    #   fetch_crypto_order_status
    #   fetch_order_queue

    # --------------- Methods for Trading --------------- #

    # Not implemented functions:
    #   order_limit
    #   order_option_limit

    # ------------- Helper methods ------------- #

    def get_data_from_polygon(
        self,
        symbol: str,
        multipler: int,
        timespan: str,
        start: dt.datetime,
        end: dt.datetime,
    ) -> pd.DataFrame:
        if self.basic and start < now() - dt.timedelta(days=365 * 2):
            debugger.warning(
                "Start time is over two years old! Only data from the past two years will be returned for basic accounts."
            )

        if timespan == "MIN":
            timespan = "minute"
        elif timespan == "HR":
            timespan = "hour"
        elif timespan == "DAY":
            timespan = "day"

        start_str = start.strftime("%Y-%m-%d")
        end_str = end.strftime("%Y-%m-%d")

        crypto = False
        if is_crypto(symbol):
            temp_symbol = "X:" + symbol[1:] + "USD"
            crypto = True

        request_form = "https://api.polygon.io/v2/aggs/ticker/{symbol}/range/{multiplier}/{timespan}/{start}/{end}?adjusted=true&sort=desc&apiKey={api_key}"
        request = request_form.format(
            symbol=temp_symbol if crypto else symbol,
            multiplier=multipler,
            timespan=timespan,
            start=start_str,
            end=end_str,
            api_key=self.config["api_key"],
        )

        response = json.load(urllib.request.urlopen(request))

        if response["status"] != "ERROR":
            df = pd.DataFrame(response["results"]).iloc[::-1]
        else:
            debugger.error(f"Request error! Returning empty dataframe. \n {response}")
            return pd.DataFrame()

        df = self._format_df(df, symbol)
        df = df.loc[start:end]

        return df

    def _format_df(self, df: pd.DataFrame, symbol: str):
        df = df.rename(
            columns={
                "t": "timestamp",
                "o": "open",
                "c": "close",
                "h": "high",
                "l": "low",
                "v": "volume",
            }
        )
        df = df[["timestamp", "open", "high", "low", "close", "volume"]].astype(float)
        df.index = pd.to_datetime(df["timestamp"], unit="ms", utc=True)
        df.drop(columns=["timestamp"], inplace=True)
        df.columns = pd.MultiIndex.from_product([[symbol], df.columns])

        return df.dropna()

    def create_secret(self, path: str) -> bool:
        import harvest.wizard as wizard

        w = wizard.Wizard()

        w.println("Hmm, looks like you haven't set up an api key for Polygon.")
        should_setup = w.get_bool("Do you want to set it up now?", default="y")

        if not should_setup:
            w.println("You can't use Polygon without an API key.")
            w.println(
                "You can set up the credentials manually, or use other streamers."
            )
            return False

        w.println("Alright! Let's get started")

        have_account = w.get_bool("Do you have a Polygon account?", default="y")
        if not have_account:
            w.println(
                "In that case you'll first need to make an account. I'll wait here, so hit Enter or Return when you've done that."
            )
            w.wait_for_input()

        api_key = w.get_string("Enter your API key")

        w.println(f"All steps are complete now 🎉. Generating {path}...")

        d = {
            "api_key": f"{api_key}",
        }

        with open(path, "w") as file:
            yml = yaml.dump(d, file)

        w.println(
            f"{path} has been created! Make sure you keep this file somewhere secure and never share it with other people."
        )

        return True



# Builtins
import datetime as dt
import time
from pathlib import Path
import yaml
import traceback
import threading
from typing import List, Dict, Any

# External libraries
import pandas as pd

# Submodule imports
from harvest.utils import *


class API:
    """
    The API class communicates with various API endpoints to perform the
    necessary operations. The Base class defines the interface for all API classes to
    extend and implement.

    Attributes
    :interval_list: A list of supported intervals.
    :exchange: The market the API trades on. Ignored if the API is not a broker.
    """

    interval_list = [
        Interval.MIN_1,
        Interval.MIN_5,
        Interval.MIN_15,
        Interval.MIN_30,
        Interval.HR_1,
        Interval.DAY_1,
    ]

    exchange = ""

    def __init__(self, path: str = None):
        """
        Performs initializations of the class, such as setting the
        timestamp and loading credentials.

        There are three API class types, 'streamer', 'broker', and 'both'. A
        'streamer' is responsible for fetching data and interacting with
        the queue to store data. A 'broker' is used solely for buying and
        selling stocks, cryptos and options. Finally, 'both' is used to
        indicate that the broker fetch data and buy and sell stocks.

        All subclass implementations should call this __init__ method
        using `super().__init__(path)`.

        :path: path to the YAML file containing credentials to communicate with the API.
            If not specified, defaults to './secret.yaml'
        """

        if path is None:
            path = "./secret.yaml"
        # Check if file exists
        yml_file = Path(path)
        if not yml_file.is_file() and not self.create_secret(path):
            debugger.debug("Broker not initalized with account information.")
        else:
            with open(path, "r") as stream:
                self.config = yaml.safe_load(stream)

        self.timestamp = now()

    def create_secret(self, path: str):
        """
        This method is called when the yaml file with credentials
        is not found."""
        # raise Exception(f"{path} was not found.")
        debugger.warning(f"Assuming API does not need account information.")
        return False

    def refresh_cred(self):
        """
        Most API endpoints, for security reasons, require a refresh of the access token
        every now and then. This method should perform a refresh of the access token.
        """
        debugger.info(f"Refreshing credentials for {type(self).__name__}.")

    def setup(self, interval: Dict, trader_main=None) -> None:
        """
        This function is called right before the algorithm begins,
        and initializes several runtime parameters like
        the symbols to watch and what interval data is needed.

        :trader_main: A callback function to the trader which will pass the data to the algorithms.
        """

        self.trader_main = trader_main

        min_interval = None
        for sym in interval:
            inter = interval[sym]["interval"]
            # If the specified interval is not supported on this API, raise Exception
            if inter < self.interval_list[0]:
                raise Exception(f"Specified interval {inter} is not supported.")
            # If the exact inteval is not supported but it can be recreated by aggregating
            # candles from a more granular interval
            if inter not in self.interval_list:
                granular_int = [i for i in self.crypto_interval_list if i < inter]
                new_inter = granular_int[-1]
                interval[sym]["aggregations"].append(inter)
                interval[sym]["interval"] = new_inter

            if min_interval is None or interval[sym]["interval"] < min_interval:
                min_interval = interval[sym]["interval"]

        self.interval = interval
        self.poll_interval = min_interval
        debugger.debug(f"Interval: {self.interval}")
        debugger.debug(f"Poll Interval: {self.poll_interval}")
        debugger.debug(f"{type(self).__name__} setup finished")

    def start(self):
        """
        This method begins streaming data from the API.

        The default implementation below is for polling the API.
        If your brokerage provides a streaming API, you should override
        this method and configure it to use that API. In that case,
        make sure to set the callback function to self.main().
        """
        cur_min = -1
        val, unit = expand_interval(self.poll_interval)

        debugger.debug(f"{type(self).__name__} started...")
        if unit == "MIN":
            sleep = val * 60 - 10
            while 1:
                cur = now()
                minutes = cur.minute
                if minutes % val == 0 and minutes != cur_min:
                    self.timestamp = cur
                    self.main()
                    time.sleep(sleep)
                cur_min = minutes
        elif unit == "HR":
            sleep = val * 3600 - 60
            while 1:
                cur = now()
                minutes = cur.minute
                if minutes == 0 and minutes != cur_min:
                    self.timestamp = cur
                    self.main()
                    time.sleep(sleep)
                cur_min = minutes
        else:
            while 1:
                cur = now()
                minutes = cur.minute
                hours = cur.hour
                if hours == 19 and minutes == 50:
                    self.timestamp = cur
                    self.main()
                    time.sleep(80000)
                cur_min = minutes

    def main(self):
        """
        This method is called at the interval specified by the user.
        It should create a dictionary where each key is the symbol for an asset,
        and the value is the corresponding data in the following pandas dataframe format:
                      Symbol
                      open   high    low close   volume
            timestamp
            ---       ---    ---     --- ---     ---

        timestamp should be an offset-aware datetime object in UTC timezone.

        The dictionary should be passed to the trader by calling `self.trader_main(dict)`
        """
        # Iterate through securities in the watchlist. For those that have
        # intervals that needs to be called now, fetch the latest data

        df_dict = {}
        for sym in self.interval:
            inter = self.interval[sym]["interval"]
            if is_freq(harvest_timestamp, inter):
                n = harvest_timestamp
                latest = self.fetch_price_history(
                    sym, inter, n - interval_to_timedelta(inter) * 2, n
                )
                debugger.debug(f"{sym} price fetch returned: {latest}")
                if latest is None or latest.empty:
                    continue
                df_dict[sym] = latest.iloc[-1]

        self.trader_main(df_dict)

    def exit(self):
        """
        This function is called after every invocation of algo's handler.
        The intended purpose is for brokers to clear any cache it may have created.
        """
        debugger.debug(f"{type(self).__name__} exited")

    def _exception_handler(func):
        """
        Wrapper to handle unexpected errors in the wrapped function.
        Most functions should be wrapped with this to properly handle errors, such as
        when internet connection is lost.

        :func: Function to wrap.
        :returns: The returned value of func if func runs properly. Raises an Exception if func fails.
        """

        def wrapper(*args, **kwargs):
            tries = 3
            while tries > 0:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    self = args[0]
                    debugger.error(f"Error: {e}")
                    traceback.print_exc()
                    debugger.error("Logging out and back in...")
                    args[0].refresh_cred()
                    tries -= 1
                    debugger.error("Retrying...")
                    continue

        return wrapper

    def _run_once(func):
        """
        Wrapper to only allows wrapped functions to be run once.

        :func: Function to wrap.
        :returns: The return of the inputted function if it has not been run before and None otherwise.
        """

        ran = False

        def wrapper(*args, **kwargs):
            nonlocal ran
            if not ran:
                ran = True
                return func(*args, **kwargs)
            return None

        return wrapper

    # -------------- Streamer methods -------------- #

    def get_current_time(self):
        return now()

    def fetch_price_history(
        self,
        symbol: str,
        interval: Interval,
        start: dt.datetime = None,
        end: dt.datetime = None,
    ) -> pd.DataFrame:
        """
        Fetches historical price data for the specified asset and period
        using the API. The first row is the earliest entry and the last
        row is the latest entry.

        :param symbol: The stock/crypto to get data for.
        :param interval: The interval of requested historical data.
        :param start: The starting date of the period, inclusive.
        :param end: The ending date of the period, inclusive.
        :returns: A pandas dataframe, same format as main()
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this streamer method: `fetch_price_history`."
        )

    def fetch_chain_info(self, symbol: str):
        """
        Returns information about the symbol's options

        :param symbol: Stock symbol. Cannot use crypto.
        :returns: A dict with the following keys and values:
            - id: ID of the option chain
            - exp_dates: List of expiration dates as datetime objects
            - multiplier: Multiplier of the option, usually 100
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this streamer method: `fetch_chain_info`."
        )

    def fetch_chain_data(self, symbol: str):
        """
        Returns the option chain for the specified symbol.

        :param symbol: Stock symbol. Cannot use crypto.
        :returns: A dataframe in the following format:

                    exp_date strike  type
            OCC
            ---     ---      ---     ---
        exp_date should be a timezone-aware datetime object localized to UTC
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this streamer method: `fetch_chain_data`."
        )

    def fetch_option_market_data(self, symbol: str):
        """
        Retrieves data of specified option.

        :param symbol:    OCC symbol of option
        :returns:   A dictionary:
            - price: price of option
            - ask: ask price
            - bid: bid price
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this streamer method: `fetch_option_market_data`."
        )

    def fetch_market_hours(self, date: datetime.date):
        """
        Returns the market hours for a given day.
        Hours are based on the exchange specified in the class's 'exchange' attribute.

        :returns: A dictionary with the following keys and values:
            - open: Time the market opens in UTC timezone.
            - close: Time the market closes in UTC timezone.
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `fetch_market_hours`."
        )

    # ------------- Broker methods ------------- #

    def fetch_stock_positions(self):
        """
        Returns all current stock positions

        :returns: A list of dictionaries with the following keys and values:
            - symbol: Ticker symbol of the stock
            - avg_price: The average price the stock was bought at
            - quantity: Quantity owned
        """
        debugger.error(
            f"{type(self).__name__} does not support this broker method: `fetch_stock_positions`. Returning an empty list."
        )
        return []

    def fetch_option_positions(self):
        """
        Returns all current option positions

        :returns: A list of dictionaries with the following keys and values:
            - symbol: OCC symbol of the option
            - base_symbol: Ticker symbol of the underlying stock
            - avg_price: Average price the option was bought at
            - quantity: Quantity owned
            - multiplier: How many stocks each option represents
            - exp_date: When the option expires
            - strike_price: Strike price of the option
            - type: 'call' or 'put'
        """
        debugger.error(
            f"{type(self).__name__} does not support this broker method: `fetch_option_positions`. Returning an empty list."
        )
        return []

    def fetch_crypto_positions(self):
        """
        Returns all current crypto positions

        :returns: A list of dictionaries with the following keys and values:
            - symbol: Ticker symbol for the crypto, prepended with an '@'
            - avg_price: The average price the crypto was bought at
            - quantity: Quantity owned
        """
        debugger.error(
            f"{type(self).__name__} does not support this broker method: `fetch_crypto_positions`. Returning an empty list."
        )
        return []

    def update_option_positions(self, positions: List[Any]):
        """
        Updates entries in option_positions list with the latest option price.
        This is needed as options are priced based on various metrics,
        and cannot be easily calculated from stock prices.

        :positions: The option_positions list in the Trader class.
        :returns: Nothing
        """
        debugger.error(
            f"{type(self).__name__} does not support this broker method: `update_option_positions`. Doing nothing."
        )

    def fetch_account(self):
        """
        Returns current account information from the brokerage.

        :returns: A dictionary with the following keys and values:
            - equity: Total assets in the brokerage
            - cash: Total cash in the brokerage
            - buying_power: Total buying power
            - multiplier: Scale of leverage, if leveraging
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `fetch_account`."
        )

    def fetch_stock_order_status(self, id):
        """
        Returns the status of a stock order with the given id.

        :id: ID of the stock order

        :returns: A dictionary with the following keys and values:
            - type: 'STOCK'
            - id: ID of the order
            - symbol: Ticker of stock
            - quantity: Quantity ordered
            - filled_quantity: Quantity filled so far
            - side: 'buy' or 'sell'
            - time_in_force: Time the order is in force
            - status: Status of the order
            - filled_time: Time the order was filled
            - filled_price: Price the order was filled at
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `fetch_stock_order_status`."
        )

    def fetch_option_order_status(self, id):
        """
        Returns the status of a option order with the given id.

        :id: ID of the option order

        :returns: A dictionary with the following keys and values:
            - type: 'OPTION'
            - id: ID of the order
            - symbol: OCC symbol of option
            - quantity: Quantity ordered
            - filled_quantity: Quantity filled so far
            - side: 'buy' or 'sell'
            - time_in_force: Time the order is in force
            - status: Status of the order
            - filled_time: Time the order was filled
            - filled_price: Price the order was filled at
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `fetch_option_order_status`."
        )

    def fetch_crypto_order_status(self, id):
        """
        Returns the status of a crypto order with the given id.

        :id: ID of the crypto order

        :returns: A dictionary with the following keys and values:
            - type: 'CRYPTO'
            - id: ID of the order
            - symbol: Ticker of crypto
            - quantity: Quantity ordered
            - filled_quantity: Quantity filled so far
            - side: 'buy' or 'sell'
            - time_in_force: Time the order is in force
            - status: Status of the order
            - filled_time: Time the order was filled
            - filled_price: Price the order was filled at
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `fetch_crypto_order_status`."
        )

    def fetch_order_queue(self):
        """
        Returns all current pending orders

        returns: A list of dictionaries with the following keys and values:
            For stocks and crypto:
                - type: "STOCK" or "CRYPTO"
                - symbol: Symbol of asset
                - quantity: Quantity ordered
                - filled_qty: Quantity filled
                - id: ID of order
                - time_in_force: Time in force
                - status: Status of the order
                - side: 'buy' or 'sell'
                - filled_time: Time the order was filled
                - filled_price: Price the order was filled at
            For options:
                - type: "OPTION",
                - symbol: OCC symbol of option
                - base_symbol:
                - quantity: Quantity ordered
                - filled_qty: Quantity filled
                - filled_time: Time the order was filled
                - filled_price: Price the order was filled at
                - id: ID of order
                - time_in_force: Time in force
                - status: Status of the order
                - legs: A list of dictionaries with keys:
                    - id: id of leg
                    - side: 'buy' or 'sell'
                    Harvest does not support buying multiple options in a single transaction,
                    so legs will always have a length of 1.

        """
        debugger.error(
            f"{type(self).__name__} does not support this broker method: `fetch_order_queue`. Returning an empty list."
        )
        return []

    # --------------- Methods for Trading --------------- #

    def order_stock_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        """
        Places a limit order.

        :symbol:    symbol of stock
        :side:      'buy' or 'sell'
        :quantity:  quantity to buy or sell
        :limit_price:   limit price
        :in_force:  'gtc' by default
        :extended:  'False' by default

        :returns: A dictionary with the following keys and values:
            - id: ID of order
            - symbol: symbol of asset
            Raises an exception if order fails.
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `order_stock_limit`."
        )

    def order_crypto_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        """
        Places a limit order.

        :symbol:    symbol of crypto
        :side:      'buy' or 'sell'
        :quantity:  quantity to buy or sell
        :limit_price:   limit price
        :in_force:  'gtc' by default
        :extended:  'False' by default

        :returns: A dictionary with the following keys and values:
            - id: ID of order
            - symbol: symbol of asset
            Raises an exception if order fails.
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `order_crypto_limit`."
        )

    def order_option_limit(
        self,
        side: str,
        symbol: str,
        quantity: float,
        limit_price: float,
        option_type: str,
        exp_date: dt.datetime,
        strike: float,
        in_force: str = "gtc",
    ):
        """
        Order an option.

        :side:      'buy' or 'sell'
        :symbol:    symbol of asset
        :in_force:
        :limit_price: limit price
        :quantity:  quantity to sell or buy
        :exp_date:  expiration date
        :strike:    strike price
        :option_type:      'call' or 'put'

        :returns: A dictionary with the following keys and values:
            - type: 'OPTION'
            - id: ID of order
            - symbol: symbol of asset
            Raises an exception if order fails.
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support this broker method: `order_option_limit`."
        )

    # -------------- Built-in methods -------------- #
    # These do not need to be re-implemented in a subclass

    def buy(
        self,
        symbol: str,
        quantity: int,
        limit_price: float,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        """
        Buys the specified asset.

        :symbol:    Symbol of the asset to buy
        :quantity:  Quantity of asset to buy
        :limit_price:   Limit price to buy at
        :in_force:  Duration the order is in force
        :extended:  Whether to trade in extended hours or not.

        :returns: The result of order_limit(). Returns None if there is an issue with the parameters.
        """

        debugger.debug(f"{type(self).__name__} ordered a buy of {quantity} {symbol}")
        typ = symbol_type(symbol)
        if typ == "STOCK":
            return self.order_stock_limit(
                "buy", symbol, quantity, limit_price, in_force, extended
            )
        elif typ == "CRYPTO":
            return self.order_crypto_limit(
                "buy", symbol[1:], quantity, limit_price, in_force, extended
            )
        elif typ == "OPTION":
            sym, exp_date, option_type, strike = self.occ_to_data(symbol)
            return self.order_option_limit(
                "buy",
                sym,
                quantity,
                limit_price,
                option_type,
                exp_date,
                strike,
                in_force,
            )
        else:
            debugger.error(f"Invalid asset type for {symbol}")

    def sell(
        self,
        symbol: str = None,
        quantity: int = 0,
        limit_price: float = 0.0,
        in_force: str = "gtc",
        extended: bool = False,
    ):
        """Sells the specified asset.

        :symbol:    Symbol of the asset to buy
        :quantity:  Quantity of asset to buy
        :limit_price:   Limit price to buy at
        :in_force:  Duration the order is in force
        :extended:  Whether to trade in extended hours or not.

        :returns: The result of order_limit(). Returns None if there is an issue with the parameters.
        """

        debugger.debug(f"{type(self).__name__} ordered a sell of {quantity} {symbol}")

        typ = symbol_type(symbol)
        if typ == "STOCK":
            return self.order_stock_limit(
                "sell", symbol, quantity, limit_price, in_force, extended
            )
        elif typ == "CRYPTO":
            return self.order_crypto_limit(
                "sell", symbol[1:], quantity, limit_price, in_force, extended
            )
        elif typ == "OPTION":
            sym, exp_date, option_type, strike = self.occ_to_data(symbol)
            return self.order_option_limit(
                "sell",
                sym,
                quantity,
                limit_price,
                option_type,
                exp_date,
                strike,
                in_force,
            )
        else:
            debugger.error(f"Invalid asset type for {symbol}")

    # def buy_option(self, symbol: str, quantity: int = 0, in_force: str = "gtc"):
    #     """
    #     Buys the specified option.

    #     :symbol:    Symbol of the asset to buy, in OCC format.
    #     :quantity:  Quantity of asset to buy
    #     :in_force:  Duration the order is in force

    #     :returns: The result of order_option_limit(). Returns None if there is an issue with the parameters.
    #     """
    #     if quantity <= 0.0:
    #         debugger.error(
    #             f"Quantity cannot be less than or equal to 0: was given {quantity}"
    #         )
    #         return None
    #     if self.trader is None:
    #         buy_power = self.fetch_account()["buying_power"]
    #         price = self.streamer.fetch_option_market_data(symbol)["price"]
    #     else:
    #         buy_power = self.trader.account["buying_power"]
    #         price = self.trader.streamer.fetch_option_market_data(symbol)["price"]

    #     limit_price = mark_up(price)
    #     total_price = limit_price * quantity

    #     if total_price >= buy_power:
    #         debugger.warning(
    #             "Not enough buying power.\n" +
    #             f"Total price ({price} * {quantity} * 1.05 = {limit_price*quantity}) exceeds buying power {buy_power}.\n" +
    #             "Reduce purchase quantity or increase buying power."
    #         )

    #     sym, date, option_type, strike = self.occ_to_data(symbol)
    #     return self.order_option_limit(
    #         "buy",
    #         sym,
    #         quantity,
    #         limit_price,
    #         option_type,
    #         date,
    #         strike,
    #         in_force=in_force,
    #     )

    # def sell_option(self, symbol: str, quantity: int = 0, in_force: str = "gtc"):
    #     """
    #     Sells the specified option.

    #     :symbol:    Symbol of the asset to buy, in OCC format.
    #     :quantity:  Quantity of asset to buy
    #     :in_force:  Duration the order is in force

    #     :returns: The result of order_option_limit(). Returns None if there is an issue with the parameters.
    #     """
    #     if quantity <= 0.0:
    #         debugger.error(
    #             f"Quantity cannot be less than or equal to 0: was given {quantity}"
    #         )
    #         return None
    #     if self.trader is None:
    #         price = self.streamer.fetch_option_market_data(symbol)["price"]
    #     else:
    #         price = self.trader.streamer.fetch_option_market_data(symbol)["price"]

    #     limit_price = mark_down(price)

    #     debugger.debug(f"{type(self).__name__} ordered a sell of {quantity} {symbol}")
    #     sym, date, option_type, strike = self.occ_to_data(symbol)
    #     return self.order_option_limit(
    #         "sell",
    #         sym,
    #         quantity,
    #         limit_price,
    #         option_type,
    #         date,
    #         strike,
    #         in_force=in_force,
    #     )

    # -------------- Helper methods -------------- #

    def has_interval(self, interval: str):
        return interval in self.interval_list

    def data_to_occ(
        self, symbol: str, date: dt.datetime, option_type: str, price: float
    ):
        """
        Converts data into a OCC format string
        """
        occ = symbol + ((6 - len(symbol)) * " ")
        occ = occ + date.strftime("%y%m%d")
        occ = occ + "C" if option_type == "call" else occ + "P"
        occ = occ + f"{int(price*1000):08}"
        return occ

    def occ_to_data(self, symbol: str):
        original_symbol = symbol
        debugger.debug(f"Converting {symbol} to data")
        try:
            sym = ""
            symbol = symbol.replace(" ", "")
            i = 0
            while symbol[i].isalpha():
                i += 1
            sym = symbol[0:i]
            symbol = symbol[i:]
            debugger.debug(f"{sym}, {symbol}")

            date = dt.datetime.strptime(symbol[0:6], "%y%m%d")
            debugger.debug(f"{date}, {symbol}")
            option_type = "call" if symbol[6] == "C" else "put"
            debugger.debug(f"{option_type}, {symbol}")
            price = float(symbol[7:]) / 1000
            debugger.debug(f"{price}, {symbol}")
            return sym, date, option_type, price
        except Exception as e:
            debugger.error(f"Error parsing OCC symbol: {original_symbol}, {e}")
            # return None, None, None, None
            raise Exception(f"Error parsing OCC symbol: {original_symbol}, {e}")

    def current_timestamp(self):
        return self.timestamp


class StreamAPI(API):
    """ """

    def __init__(self, path: str = None):
        super().__init__(path)

        self.block_lock = (
            threading.Lock()
        )  # Lock for streams that receive data asynchronously.
        self.block_queue = {}
        self.first = True

    def setup(self, interval: Dict, trader_main=None) -> None:
        super().setup(interval, trader_main)
        self.blocker = {}

    def start(self):
        debugger.debug(f"{type(self).__name__} started...")

    def main(self, df_dict):
        """
        Streaming is event driven, so sometimes not all data comes in at once.
        StreamAPI class
        """
        self.block_lock.acquire()
        got = [k for k in df_dict]
        # First, identify which symbols need to have data fetched
        # for this timestamp
        if self.first:
            self.needed = [
                sym
                for sym in self.interval
                if is_freq(now(), self.interval[sym]["interval"])
            ]
            self.timestamp = df_dict[got[0]].index[0]

        debugger.debug(f"Needs: {self.needed}")
        debugger.debug(f"Got data for: {got}")
        missing = list(set(self.needed) - set(got))
        debugger.debug(f"Still need data for: {missing}")

        self.block_queue.update(df_dict)
        # debugger.debug(self.block_queue)

        # If all data has been received, pass on the data
        if len(missing) == 0:
            debugger.debug("All data received")
            self.trader_main(self.block_queue)
            self.block_queue = {}
            self.all_recv = True
            self.first = True
            self.block_lock.release()
            return

        # If there are data that has not been received, start a timer
        if self.first:
            timer = threading.Thread(target=self.timeout, daemon=True)
            timer.start()
            self.all_recv = False
            self.first = False

        self.needed = missing
        self.got = got
        self.block_lock.release()

    def timeout(self):
        debugger.debug("Begin timeout timer")
        time.sleep(1)
        if not self.all_recv:
            debugger.debug("Force flush")
            self.flush()

    def flush(self):
        # For missing data, return a OHLC with all zeroes.
        self.block_lock.acquire()
        for n in self.needed:
            data = pd.DataFrame(
                {"open": 0, "high": 0, "low": 0, "close": 0, "volume": 0},
                index=[self.timestamp],
            )

            data.columns = pd.MultiIndex.from_product([[n], data.columns])
            self.block_queue[n] = data
        self.block_lock.release()
        self.trader_main(self.block_queue)
        self.block_queue = {}
